import sys
from abc import ABC, abstractmethod
from fnmatch import fnmatch
from typing import Any, Callable, Dict, Generic, Tuple, Type, TypeVar, Union

if sys.version_info >= (3, 11):
    from typing import Unpack
else:
    from typing_extensions import Unpack

from kloppy.domain import Code, DataRecord, DatasetType, Event, Frame
from kloppy.domain.services.transformers.attribute import (
    DefaultCodeTransformer,
    DefaultEventTransformer,
    DefaultFrameTransformer,
)
from kloppy.exceptions import KloppyError

T = TypeVar("T", bound=DataRecord)
Column = Union[str, Callable[[T], Any]]
NamedColumns = Dict[str, Column]


class DataRecordToDictTransformer(ABC, Generic[T]):
    @abstractmethod
    def default_transformer(self) -> Callable[[T], Dict]:
        ...

    def __init__(
        self,
        *columns: Unpack[Tuple[Column]],
        **named_columns: NamedColumns,
    ):
        if not columns and not named_columns:
            converter = self.default_transformer()
        else:
            default = self.default_transformer()
            has_string_columns = any(
                not callable(column) for column in columns
            )

            def converter(data_record: T) -> Dict[str, Any]:
                if has_string_columns:
                    default_row = default(data_record)
                else:
                    default_row = {}

                row = {}
                for column in columns:
                    if callable(column):
                        res = column(data_record)
                        if not isinstance(res, dict):
                            raise KloppyError(
                                "A function column should return a dictionary"
                            )
                        row.update(res)
                    else:
                        if column == "*":
                            row.update(default_row)
                        elif "*" in column:
                            row.update(
                                {
                                    k: v
                                    for k, v in default_row.items()
                                    if fnmatch(k, column)
                                }
                            )
                        elif column in default_row:
                            row[column] = default_row[column]
                        else:
                            row[column] = getattr(data_record, column, None)

                for name, column in named_columns.items():
                    row[name] = (
                        column(data_record) if callable(column) else column
                    )

                return row

        self.converter = converter

    def __call__(self, data_record: T) -> Dict[str, Any]:
        return self.converter(data_record)


class EventToDictTransformer(DataRecordToDictTransformer[Event]):
    def default_transformer(self) -> Callable[[Event], Dict]:
        return DefaultEventTransformer()


class FrameToDictTransformer(DataRecordToDictTransformer[Frame]):
    def default_transformer(self) -> Callable[[Frame], Dict]:
        return DefaultFrameTransformer()


class CodeToDictTransformer(DataRecordToDictTransformer[Code]):
    def default_transformer(self) -> Callable[[Code], Dict]:
        return DefaultCodeTransformer()


def get_transformer_cls(
    dataset_type: DatasetType,
) -> Type[DataRecordToDictTransformer]:
    if dataset_type == DatasetType.EVENT:
        return EventToDictTransformer
    elif dataset_type == DatasetType.TRACKING:
        return FrameToDictTransformer
    elif dataset_type == DatasetType.CODE:
        return CodeToDictTransformer



from datetime import datetime, timedelta, timezone
from pathlib import Path

import pytest

from kloppy import statsperform
from kloppy.domain import (
    DatasetFlag,
    EventDataset,
    OptaCoordinateSystem,
    Orientation,
    PassResult,
    Point,
    Point3D,
    PositionType,
    Provider,
    SportVUCoordinateSystem,
    Time,
    TrackingDataset,
)


@pytest.fixture(scope="module")
def event_metadata_xml(base_dir: Path) -> Path:
    return base_dir / "files" / "statsperform_event_ma1.xml"


@pytest.fixture(scope="module")
def event_metadata_json(base_dir: Path) -> Path:
    return base_dir / "files" / "statsperform_event_ma1.json"


@pytest.fixture(scope="module")
def event_data_xml(base_dir: Path) -> Path:
    return base_dir / "files" / "statsperform_event_ma3.xml"


@pytest.fixture(scope="module")
def event_data_json(base_dir: Path) -> Path:
    return base_dir / "files" / "statsperform_event_ma3.json"


@pytest.fixture(scope="module")
def tracking_metadata_xml(base_dir: Path) -> Path:
    return base_dir / "files" / "statsperform_tracking_ma1.xml"


@pytest.fixture(scope="module")
def tracking_metadata_json(base_dir: Path) -> Path:
    return base_dir / "files" / "statsperform_tracking_ma1.json"


@pytest.fixture(scope="module")
def tracking_data(base_dir: Path) -> Path:
    return base_dir / "files" / "statsperform_tracking_ma25.txt"


@pytest.fixture(scope="module", params=["xml", "json"])
def tracking_dataset(
    request: pytest.FixtureRequest,
    tracking_metadata_xml: Path,
    tracking_metadata_json: Path,
    tracking_data: Path,
) -> TrackingDataset:
    return statsperform.load_tracking(
        ma1_data=(
            tracking_metadata_xml
            if request.param == "xml"
            else tracking_metadata_json
        ),
        ma25_data=tracking_data,
        tracking_system="sportvu",
        only_alive=False,
        coordinates="sportvu",
    )


@pytest.fixture(scope="module", params=["xml", "json"])
def event_dataset(
    request: pytest.FixtureRequest,
    event_metadata_xml: Path,
    event_metadata_json: Path,
    event_data_xml: Path,
    event_data_json: Path,
) -> EventDataset:
    return statsperform.load_event(
        ma1_data=(
            event_metadata_xml
            if request.param == "xml"
            else event_metadata_json
        ),
        ma3_data=event_data_xml if request.param == "xml" else event_data_json,
        coordinates="opta",
    )


class TestStatsPerformMetadata:
    """Tests related to deserializing the MA1 meta data feed."""

    def test_provider(self, tracking_dataset: TrackingDataset):
        assert tracking_dataset.metadata.provider == Provider.STATSPERFORM

    def test_orientation(self, event_dataset: EventDataset):
        assert event_dataset.metadata.coordinate_system == OptaCoordinateSystem(
            # StatsPerform does not provide pitch dimensions
            pitch_length=None,
            pitch_width=None,
        )

    def test_teams(self, tracking_dataset: TrackingDataset):
        home_team = tracking_dataset.metadata.teams[0]
        home_player = home_team.players[2]

        assert home_player.player_id == "5g5wwp5luxo1rz1kp6chvz0x6"
        assert tracking_dataset.records[0].players_coordinates[
            home_player
        ] == Point(x=68.689, y=39.75)
        assert home_player.starting_position == PositionType.LeftCenterBack
        assert home_player.jersey_no == 32
        assert home_player.starting
        assert home_player.team == home_team

        away_team = tracking_dataset.metadata.teams[1]
        away_player = away_team.players[3]
        assert away_player.player_id == "72d5uxwcmvhd6mzthxuvev1sl"
        assert tracking_dataset.records[0].players_coordinates[
            away_player
        ] == Point(x=30.595, y=44.022)
        assert away_player.starting_position == PositionType.RightCenterBack
        assert away_player.jersey_no == 2
        assert away_player.starting
        assert away_player.team == away_team

        away_substitute = away_team.players[15]
        assert away_substitute.jersey_no == 18
        assert away_substitute.starting_position is None
        assert not away_substitute.starting
        assert away_substitute.team == away_team

        home_gk = home_team.get_player_by_id("6wfwy94p5bm0zv3aku0urfq39")
        assert home_gk.first_name == "Benjamin Pascal"
        assert home_gk.last_name == "Lecomte"

    def test_periods(self, tracking_dataset: TrackingDataset):
        assert len(tracking_dataset.metadata.periods) == 2
        assert tracking_dataset.metadata.periods[0].id == 1
        assert tracking_dataset.metadata.periods[
            0
        ].start_timestamp == datetime(
            2020, 8, 23, 11, 0, 10, tzinfo=timezone.utc
        )
        assert tracking_dataset.metadata.periods[0].end_timestamp == datetime(
            2020, 8, 23, 11, 48, 15, tzinfo=timezone.utc
        )

        assert tracking_dataset.metadata.periods[1].id == 2
        assert tracking_dataset.metadata.periods[
            1
        ].start_timestamp == datetime(
            2020, 8, 23, 12, 6, 22, tzinfo=timezone.utc
        )
        assert tracking_dataset.metadata.periods[1].end_timestamp == datetime(
            2020, 8, 23, 12, 56, 30, tzinfo=timezone.utc
        )

    def test_enriched_metadata(self, tracking_dataset: TrackingDataset):
        date = tracking_dataset.metadata.date
        if date:
            assert isinstance(date, datetime)
            assert date == datetime(2020, 8, 23, 0, 0, tzinfo=timezone.utc)

        game_week = tracking_dataset.metadata.game_week
        if game_week:
            assert isinstance(game_week, str)
            assert game_week == "1"

        game_id = tracking_dataset.metadata.game_id
        if game_id:
            assert isinstance(game_id, str)
            assert game_id == "7ijuqohwgmplbxdj1625sxwfe"


class TestStatsPerformEvent:
    """Tests related to deserializing the MA3 event data feed.

    See Also:
        kloppy.tests.test_opta.TestOptaEvent
    """

    def test_deserialize_all(self, event_dataset: EventDataset):
        assert event_dataset.metadata.provider == Provider.STATSPERFORM

        assert len(event_dataset.records) == 1643

        substitution_events = event_dataset.find_all("substitution")
        assert len(substitution_events) == 9

        m_wintzheimer = event_dataset.metadata.teams[0].get_player_by_id(
            "aksjicf4keobpav3tuujngell"
        )
        b_jatta = event_dataset.metadata.teams[0].get_player_by_id(
            "3mp7p8tytgkbwi8itxl5mfkrt"
        )

        first_sub = substitution_events[0]

        assert first_sub.time == Time(
            period=event_dataset.metadata.periods[1],
            timestamp=timedelta(seconds=946, microseconds=475000),
        )
        assert first_sub.player == m_wintzheimer
        assert first_sub.replacement_player == b_jatta

    def test_pass_receiver(self, event_dataset: EventDataset):
        """It should impute the intended receiver of a pass."""
        # Completed passes should get a receiver
        complete_pass = event_dataset.get_event_by_id("2328589789")
        assert (
            complete_pass.receiver_player.player_id
            == "apdrig6xt1hxub1986s3uh1x"
        )
        assert (
            complete_pass.receive_timestamp
            == complete_pass.next_record.timestamp
        )

        # When a pass is challenged the receipt is the next next event
        challenged_pass = event_dataset.get_event_by_id("2328600271")
        assert (
            challenged_pass.receiver_player.player_id
            == "ci4pwzieoc94uj3i1371bsatx"
        )
        assert (
            challenged_pass.receive_timestamp
            == challenged_pass.next_record.next_record.timestamp
        )

        # Passes should be received within 30 seconds
        assert all(
            [
                p.receive_timestamp.total_seconds()
                - p.timestamp.total_seconds()
                < 30
                for p in event_dataset.find_all("pass")
                if p.receive_timestamp is not None
            ]
        )

        # Passes that result in a loss of possession should not have a receiver
        turnover_passes = [
            p
            for p in event_dataset.find_all("pass")
            if p.next_record
            and p.ball_owning_team != p.next_record.ball_owning_team
        ]
        assert all(p.receiver_player is None for p in turnover_passes)

        # Failed passes should not have a receiver
        failed_pass = event_dataset.get_event_by_id("2328591011")
        assert failed_pass.receiver_player is None
        out_pass = event_dataset.get_event_by_id("2328590733")
        assert out_pass.receiver_player is None

        # When a pass is interrupted by a foul the receiver is not set
        fouled_pass = event_dataset.get_event_by_id("2328589929")
        assert fouled_pass.receiver_player is None

        # Deflected passes should not have a receiver
        deflected_pass = event_dataset.get_event_by_id("2328596237")
        assert deflected_pass.receiver_player is None


class TestStatsPerformTracking:
    """Tests related to deserializing tracking data delivered by StatsPerform."""

    def test_orientation(self, tracking_dataset: TrackingDataset):
        assert tracking_dataset.metadata.orientation == Orientation.AWAY_HOME

    def test_framerate(self, tracking_dataset: TrackingDataset):
        assert tracking_dataset.metadata.frame_rate == 10.0

    def test_flags(self, tracking_dataset):
        assert tracking_dataset.metadata.flags == DatasetFlag.BALL_STATE

    def test_correct_deserialization_limit_sample(
        self, tracking_data: Path, tracking_metadata_xml: Path
    ):

        tracking_dataset = statsperform.load_tracking(
            ma1_data=tracking_metadata_xml,
            ma25_data=tracking_data,
            tracking_system="sportvu",
            coordinates="sportvu",
            pitch_length=105,
            pitch_width=68,
            limit=50,
        )
        assert len(tracking_dataset.records) == 50

        tracking_dataset = statsperform.load_tracking(
            ma1_data=tracking_metadata_xml,
            ma25_data=tracking_data,
            tracking_system="sportvu",
            coordinates="sportvu",
            pitch_length=105,
            pitch_width=68,
            limit=25,
            sample_rate=(1 / 2),
        )
        assert len(tracking_dataset.records) == 25

    def test_coordinate_system_without_pitch_dimensions(
        self, tracking_data: Path, tracking_metadata_xml: Path
    ):
        tracking_dataset = statsperform.load_tracking(
            ma1_data=tracking_metadata_xml,
            ma25_data=tracking_data,
            tracking_system="sportvu",
            coordinates="sportvu",
        )
        coordinate_system = tracking_dataset.metadata.coordinate_system
        pitch_dimensions = tracking_dataset.metadata.pitch_dimensions
        assert coordinate_system == SportVUCoordinateSystem(
            # StatsPerform does not provide pitch dimensions
            pitch_length=None,
            pitch_width=None,
        )
        assert pitch_dimensions.x_dim.min == 0
        assert pitch_dimensions.x_dim.max is None
        assert pitch_dimensions.y_dim.min == 0
        assert pitch_dimensions.y_dim.max is None

    def test_coordinate_system_with_pitch_dimensions(
        self, tracking_data: Path, tracking_metadata_xml: Path
    ):
        tracking_dataset = statsperform.load_tracking(
            ma1_data=tracking_metadata_xml,
            ma25_data=tracking_data,
            tracking_system="sportvu",
            coordinates="sportvu",
            pitch_length=105,
            pitch_width=68,
        )
        coordinate_system = tracking_dataset.metadata.coordinate_system
        pitch_dimensions = tracking_dataset.metadata.pitch_dimensions
        assert coordinate_system == SportVUCoordinateSystem(
            # StatsPerform does not provide pitch dimensions
            pitch_length=105,
            pitch_width=68,
        )
        assert pitch_dimensions.x_dim.min == 0
        assert pitch_dimensions.x_dim.max == 105
        assert pitch_dimensions.y_dim.min == 0
        assert pitch_dimensions.y_dim.max == 68

    def test_deserialize_all(self, tracking_dataset: TrackingDataset):
        assert len(tracking_dataset.records) == 92

    def test_deserialize_only_alive(
        self, tracking_data: Path, tracking_metadata_xml: Path
    ):
        tracking_dataset = statsperform.load_tracking(
            ma1_data=tracking_metadata_xml,
            ma25_data=tracking_data,
            tracking_system="sportvu",
            only_alive=True,
            coordinates="sportvu",
        )
        assert len(tracking_dataset.records) == 91

    def test_timestamps(self, tracking_dataset: TrackingDataset):
        assert tracking_dataset.records[0].timestamp == timedelta(
            seconds=0
        )  # First frame
        assert tracking_dataset.records[20].timestamp == timedelta(
            seconds=2.0
        )  # Later frame
        assert tracking_dataset.records[26].timestamp == timedelta(
            seconds=0
        )  # Second period

    def test_ball_coordinates(self, tracking_dataset: TrackingDataset):
        assert tracking_dataset.records[1].ball_coordinates == Point3D(
            x=50.615, y=35.325, z=0.0
        )

    def test_player_coordinates(self, tracking_dataset: TrackingDataset):
        home_player = tracking_dataset.metadata.teams[0].players[2]
        assert tracking_dataset.records[0].players_coordinates[
            home_player
        ] == Point(x=68.689, y=39.750)

    def test_correct_normalized_deserialization(
        self, tracking_data: Path, tracking_metadata_xml: Path
    ):
        tracking_dataset = statsperform.load_tracking(
            ma1_data=tracking_metadata_xml,
            ma25_data=tracking_data,
            tracking_system="sportvu",
            pitch_length=105,
            pitch_width=68,
            only_alive=False,
            coordinates="kloppy",
        )

        assert tracking_dataset.records[1].ball_coordinates == Point3D(
            x=50.615 / 105, y=35.325 / 68, z=0.0
        )

        # Check normalised pitch dimensions
        pitch_dimensions = tracking_dataset.metadata.pitch_dimensions
        assert pitch_dimensions.x_dim.min == 0.0
        assert pitch_dimensions.x_dim.max == 1.0
        assert pitch_dimensions.y_dim.min == 0.0
        assert pitch_dimensions.y_dim.max == 1.0

        # Pitch dimensions are required to transform coordinates
        with pytest.warns(
            UserWarning,
            match="The pitch dimensions are required to transform coordinates *",
        ):
            statsperform.load_tracking(
                ma1_data=tracking_metadata_xml,
                ma25_data=tracking_data,
                tracking_system="sportvu",
                coordinates="kloppy",
            )



from collections import defaultdict, Counter
from typing import List, Dict

from kloppy.domain import AttackingDirection, Frame, Ground, Period

from .event_factory import EventFactory, create_event
from .transformers import DatasetTransformer, DatasetTransformerBuilder

# NOT YET: from .enrichers import TrackingPossessionEnricher


def avg(items: List[float]) -> float:
    if not items:
        return 0
    return sum(items) / len(items)


def attacking_direction_from_frame(frame: Frame) -> AttackingDirection:
    """This method should only be called for the first frame of a period."""
    avg_x_home = avg(
        [
            player_data.coordinates.x
            for player, player_data in frame.players_data.items()
            if player.team.ground == Ground.HOME
        ]
    )
    avg_x_away = avg(
        [
            player_data.coordinates.x
            for player, player_data in frame.players_data.items()
            if player.team.ground == Ground.AWAY
        ]
    )

    if avg_x_home < avg_x_away:
        return AttackingDirection.LTR
    else:
        return AttackingDirection.RTL


def attacking_directions_from_multi_frames(
    frames: List[Frame], periods: List[Period]
) -> Dict[int, AttackingDirection]:
    """
    with only partial tracking data we cannot rely on a single frame to
    infer the attacking directions as a simple average of only some players
    x-coords might not reflect the attacking direction.
    """
    attacking_directions = {}

    # Group attacking directions by period ID
    period_direction_map = defaultdict(list)
    for frame in frames:
        if len(frame.players_data) > 0:
            direction = attacking_direction_from_frame(frame)
        else:
            direction = AttackingDirection.NOT_SET
        period_direction_map[frame.period.id].append(direction)

    # Determine the most common attacking direction for each period
    for period in periods:
        period_id = period.id
        if period_id in period_direction_map:
            count = Counter(period_direction_map[period_id])
            attacking_directions[period_id] = count.most_common(1)[0][0]
        else:
            attacking_directions[period_id] = AttackingDirection.NOT_SET

    return attacking_directions


__all__ = [
    "DatasetTransformer",
    "DatasetTransformerBuilder",
    "EventFactory",
    "create_event",
    "attacking_direction_from_frame",
    "attacking_directions_from_multi_frames",
]



import warnings
from typing import Optional

from kloppy.domain import TrackingDataset
from kloppy.infra.serializers.tracking.tracab.deserializer import (
    TRACABDeserializer,
    TRACABInputs,
)
from kloppy.io import FileLike, open_as_file


def load(
    meta_data: FileLike,
    raw_data: FileLike,
    sample_rate: Optional[float] = None,
    limit: Optional[int] = None,
    coordinates: Optional[str] = None,
    only_alive: bool = True,
    file_format: Optional[str] = None,
) -> TrackingDataset:
    """
    Load TRACAB tracking data.

    Args:
        meta_data: A JSON or XML feed containing the meta data.
        raw_data: A JSON or dat feed containing the raw tracking data.
        sample_rate: Sample the data at a specific rate.
        limit: Limit the number of frames to load to the first `limit` frames.
        coordinates: The coordinate system to use.
        only_alive: Only include frames in which the game is not paused.
        file_format: Deprecated. The format will be inferred based on the file extensions.

    Returns:
        The parsed tracking data.

    Notes:
        Tracab distributes its metadata in various formats. Kloppy tries to
        infer automatically which format applies. Currently, kloppy supports
        the following formats:

        - **Flat XML structure**:

            <root>
                <GameID>13331</GameID>
                <CompetitionID>55</CompetitionID>
                ...
            </root>

        - **Hierarchical XML structure**:

            <match iId="1" ...>
                <period iId="1" iStartFrame="1848508" iEndFrame="1916408"/>
                ...
            </match>

        - **JSON structure**:

            {
                "GameID": 1,
                "CompetitionID": 1,
                "SeasonID": 2023,
                ...
            }

        If parsing fails for a supported format or you encounter an unsupported
        structure, please create an issue on the kloppy GitHub repository
        with a sample of the problematic data.
    """
    # file format is deprecated
    if file_format is not None:
        warnings.warn(
            "file_format is deprecated. This is now automatically infered.",
            DeprecationWarning,
            stacklevel=2,
        )

    deserializer = TRACABDeserializer(
        sample_rate=sample_rate,
        limit=limit,
        coordinate_system=coordinates,
        only_alive=only_alive,
    )
    with open_as_file(meta_data) as meta_data_fp, open_as_file(
        raw_data
    ) as raw_data_fp:
        return deserializer.deserialize(
            inputs=TRACABInputs(meta_data=meta_data_fp, raw_data=raw_data_fp)
        )



import json
import logging
from itertools import zip_longest
from typing import IO, NamedTuple, Optional

from kloppy.domain import (
    DatasetFlag,
    EventDataset,
    FormationType,
    Ground,
    Metadata,
    Orientation,
    Period,
    Player,
    Provider,
    Team,
)
from kloppy.exceptions import DeserializationError
from kloppy.infra.serializers.event.deserializer import EventDataDeserializer
from kloppy.utils import performance_logging

from . import specification as SB
from .helpers import parse_freeze_frame, parse_str_ts
from .specification import position_types_mapping

logger = logging.getLogger(__name__)


class StatsBombInputs(NamedTuple):
    event_data: IO[bytes]
    lineup_data: IO[bytes]
    three_sixty_data: Optional[IO[bytes]]


class StatsBombDeserializer(EventDataDeserializer[StatsBombInputs]):
    @property
    def provider(self) -> Provider:
        return Provider.STATSBOMB

    def deserialize(
        self, inputs: StatsBombInputs, additional_metadata
    ) -> EventDataset:
        # Intialize coordinate system transformer
        self.transformer = self.get_transformer()

        # Load data from JSON files
        # and determine fidelity versions for x/y coordinates
        with performance_logging("load data", logger=logger):
            (
                raw_events,
                lineups,
                three_sixty_data,
                data_version,
            ) = self.load_data(inputs)
        logger.info(
            "determined fidelity versions (Shot v%d / XY v%d)",
            data_version.shot_fidelity_version,
            data_version.xy_fidelity_version,
        )

        # Create teams and players
        with performance_logging("parse teams ans players", logger=logger):
            teams = self.create_teams_and_players(raw_events, lineups)

        # Create periods
        with performance_logging("parse periods", logger=logger):
            periods = self.create_periods(raw_events)

        # Create events
        with performance_logging("parse events", logger=logger):
            events = []
            for raw_event in raw_events.values():
                new_events = (
                    raw_event.set_version(data_version)
                    .set_refs(periods, teams, raw_events)
                    .deserialize(self.event_factory)
                )
                for event in new_events:
                    if self.should_include_event(event):
                        # Transform event to the coordinate system
                        event = self.transformer.transform_event(event)
                        events.append(event)

        metadata = Metadata(
            teams=teams,
            periods=periods,
            pitch_dimensions=self.transformer.get_to_coordinate_system().pitch_dimensions,
            frame_rate=None,
            orientation=Orientation.ACTION_EXECUTING_TEAM,
            flags=DatasetFlag.BALL_OWNING_TEAM | DatasetFlag.BALL_STATE,
            score=None,
            provider=Provider.STATSBOMB,
            coordinate_system=self.transformer.get_to_coordinate_system(),
            **additional_metadata,
        )
        dataset = EventDataset(metadata=metadata, records=events)
        for event in dataset:
            if "freeze_frame" in event.raw_event.get("shot", {}):
                event.freeze_frame = self.transformer.transform_frame(
                    parse_freeze_frame(
                        freeze_frame=event.raw_event["shot"]["freeze_frame"],
                        home_team=teams[0],
                        away_team=teams[1],
                        event=event,
                        fidelity_version=data_version.shot_fidelity_version,
                    )
                )
            if not event.freeze_frame and event.event_id in three_sixty_data:
                freeze_frame = three_sixty_data[event.event_id]
                event.freeze_frame = self.transformer.transform_frame(
                    parse_freeze_frame(
                        freeze_frame=freeze_frame["freeze_frame"],
                        home_team=teams[0],
                        away_team=teams[1],
                        event=event,
                        fidelity_version=data_version.xy_fidelity_version,
                        visible_area=freeze_frame["visible_area"],
                    )
                )
        return dataset

    def load_data(self, inputs: StatsBombInputs):
        raw_events = {}
        shot_fidelity_version, xy_fidelity_version = 1, 1
        for event in json.load(inputs.event_data):
            # load the event
            raw_events[event["id"]] = SB.event_decoder(event)
            # determine the fidelity version
            if "location" in event:
                x, y, *_ = event["location"]
                # find out if x and y are integers disguised as floats
                if not (x.is_integer() and y.is_integer()):
                    event_type = SB.EVENT_TYPE(event["type"])
                    if event_type in (SB.EVENT_TYPE.SHOT,):
                        shot_fidelity_version = 2
                    elif event_type in (
                        SB.EVENT_TYPE.CARRY,
                        SB.EVENT_TYPE.DRIBBLE,
                        SB.EVENT_TYPE.PASS,
                    ):
                        xy_fidelity_version = 2

        version = SB.Version(
            shot_fidelity_version=shot_fidelity_version,
            xy_fidelity_version=xy_fidelity_version,
        )

        lineups = json.load(inputs.lineup_data)

        three_sixty_data = (
            {
                item["event_uuid"]: item
                for item in json.load(inputs.three_sixty_data)
            }
            if inputs.three_sixty_data
            else {}
        )

        return raw_events, lineups, three_sixty_data, version

    def create_teams_and_players(self, raw_events, lineups):
        it_events = iter(raw_events.values())
        starting_xi_events = [
            next(it_events).raw_event,
            next(it_events).raw_event,
        ]

        # Determine home and away lineups
        home_lineup, away_lineup = (
            lineups
            if starting_xi_events[0]["team"]["id"] == lineups[0]["team_id"]
            else reversed(lineups)
        )

        # Create players and teams
        player_positions = {
            str(player["player"]["id"]): position_types_mapping[
                player["position"]["id"]
            ]
            for raw_event in starting_xi_events
            for player in raw_event["tactics"]["lineup"]
        }
        starting_formations = {
            raw_event["team"]["id"]: FormationType(
                "-".join(list(str(raw_event["tactics"]["formation"])))
            )
            for raw_event in starting_xi_events
        }

        def create_team(lineup, ground_type):
            team = Team(
                team_id=str(lineup["team_id"]),
                name=lineup["team_name"],
                ground=ground_type,
                starting_formation=starting_formations[lineup["team_id"]],
            )
            team.players = [
                Player(
                    player_id=str(player["player_id"]),
                    team=team,
                    name=player["player_name"],
                    jersey_no=int(player["jersey_number"]),
                    starting=str(player["player_id"]) in player_positions,
                    starting_position=player_positions.get(
                        str(player["player_id"])
                    ),
                )
                for player in lineup["lineup"]
            ]
            return team

        home_team = create_team(home_lineup, Ground.HOME)
        away_team = create_team(away_lineup, Ground.AWAY)
        return [home_team, away_team]

    def create_periods(self, raw_events):
        half_start_events = {}
        half_end_events = {}
        for event in raw_events.values():
            event_type = SB.EVENT_TYPE(event.raw_event["type"])
            period = event.raw_event["period"]

            if event_type == SB.EVENT_TYPE.HALF_START:
                half_start_events[period] = event.raw_event
            elif event_type == SB.EVENT_TYPE.HALF_END:
                half_end_events[period] = event.raw_event

        periods = []
        for start_event, end_event in zip_longest(
            half_start_events.values(), half_end_events.values()
        ):
            if start_event is None or end_event is None:
                raise DeserializationError(
                    "Failed to determine start and end time of periods."
                )
            start_timestamp = (
                periods[-1].end_timestamp
                + parse_str_ts(start_event["timestamp"])
                if len(periods) > 0
                else parse_str_ts(start_event["timestamp"])
            )
            end_timestamp = (
                start_timestamp + parse_str_ts(end_event["timestamp"])
                if end_event
                else None
            )
            period = Period(
                id=int(start_event["period"]),
                start_timestamp=start_timestamp,
                end_timestamp=end_timestamp,
            )
            periods.append(period)
        return periods



from .deserializer import (
    MetricaEPTSTrackingDataDeserializer,
    MetricaEPTSTrackingDataInputs,
)

__all__ = [
    "MetricaEPTSTrackingDataDeserializer",
    "MetricaEPTSTrackingDataInputs",
]



from datetime import timedelta
from typing import Tuple

import pytest

from kloppy import statsbomb
from kloppy.domain import Time, Period, TimeContainer


@pytest.fixture
def periods() -> Tuple[Period, Period, Period]:
    period1 = Period(
        id=1,
        start_timestamp=timedelta(seconds=0),
        end_timestamp=timedelta(seconds=2700),
    )
    period2 = Period(
        id=2,
        start_timestamp=timedelta(seconds=0),
        end_timestamp=timedelta(seconds=3000),
    )
    period3 = Period(
        id=3,
        start_timestamp=timedelta(seconds=0),
        end_timestamp=timedelta(seconds=1000),
    )
    period1.set_refs(None, period2)
    period2.set_refs(period1, period3)
    period3.set_refs(period2, None)
    return period1, period2, period3


class TestAbsTime:
    def test_subtract_timedelta_same_period(self, periods):
        """Test subtract with non-period overlapping timedelta."""
        period1, *_ = periods

        time = Time(period=period1, timestamp=timedelta(seconds=1800))

        assert time - timedelta(seconds=1000) == Time(
            period=period1, timestamp=timedelta(seconds=800)
        )

    def test_subtract_timedelta_spans_periods(self, periods):
        """Test subtract that goes over spans multiple period.


        +-------+-------+------+
        | 2700  | 3000  | 1000 |
        +-------+-------+------+
                           ^ 800

        In period 3:  800 -   3200 left -    0
        In period 2: 3000 -    200 left -    0
        In period 1:  200 -      0 left - 2500
        """
        period1, period2, period3 = periods

        time = Time(period=period3, timestamp=timedelta(seconds=800))

        assert time - timedelta(seconds=4000) == Time(
            period=period1, timestamp=timedelta(seconds=2500)
        )

    def test_subtract_timedelta_over_start(self, periods):
        """Test subtract that goes over start of first period. This should return start of match."""
        period1, *_ = periods

        time = Time(period=period1, timestamp=timedelta(seconds=1800))

        assert time - timedelta(seconds=2000) == Time(
            period=period1, timestamp=timedelta(0)
        )

    def test_subtract_two_abstime(self, periods):
        """Subtract two AbsTime in same period"""
        period1, *_ = periods
        time1 = Time(period=period1, timestamp=timedelta(seconds=1000))
        time2 = Time(period=period1, timestamp=timedelta(seconds=800))

        assert time1 - time2 == timedelta(seconds=200)

    def test_subtract_two_abstime_spans_periods(self, periods):
        """Subtract AbsTime over multiple periods."""
        period1, period2, period3 = periods
        time1 = Time(period=period1, timestamp=timedelta(seconds=800))
        time2 = Time(period=period2, timestamp=timedelta(seconds=800))

        assert time2 - time1 == timedelta(seconds=2700)

    def test_add_timedelta_same_period(self, periods):
        """Test add timedelta in same period"""
        period1, *_ = periods

        time = Time(period=period1, timestamp=timedelta(seconds=800))
        assert time + timedelta(seconds=100) == Time(
            period=period1, timestamp=timedelta(seconds=900)
        )

    def test_add_timedelta_spans_periods(self, periods):
        """
        +-------+-------+------+
        | 2700  | 3000  | 1000 |
        +-------+-------+------+
           ^ 800

        In period 3:  1900 -   3100 left
        In period 2:  3000 -    100 left
        In period 1:   100 -      0 left
        """
        period1, period2, period3 = periods

        time = Time(period=period1, timestamp=timedelta(seconds=800))
        assert time + timedelta(seconds=5000) == Time(
            period=period3, timestamp=timedelta(seconds=100)
        )

        assert time + timedelta(seconds=2600) == Time(
            period=period2, timestamp=timedelta(seconds=700)
        )

    def test_statsbomb_formation_changes(self, base_dir):
        dataset = statsbomb.load(
            lineup_data=base_dir / "files/statsbomb_lineup.json",
            event_data=base_dir / "files/statsbomb_event.json",
        )
        formation_changes = dataset.filter("formation_change")

        # Determine time until first formation change
        diff = (
            formation_changes[0].time - dataset.metadata.periods[0].start_time
        )
        assert diff == timedelta(seconds=2705.267)

        # Time until last formation change
        diff = (
            formation_changes[-1].time - dataset.metadata.periods[0].start_time
        )
        assert diff == timedelta(seconds=5067.367)

    def test_statsbomb_minuted_played(self, base_dir):
        def __period_offset(period_id, dataset):
            for period in dataset.metadata.periods:
                if period.id == period_id - 1:
                    return period.end_time.timestamp.total_seconds()
                else:
                    return 0

        dataset = statsbomb.load(
            lineup_data=base_dir / "files/statsbomb_lineup.json",
            event_data=base_dir / "files/statsbomb_event.json",
        )

        minutes_played = dataset.aggregate("minutes_played")

        home_team, away_team = dataset.metadata.teams

        minutes_played_map = {
            item.player: item.duration for item in minutes_played
        }

        """
        3109 - 0:00:00.000000 - Malcom
        3501 - 0:47:32.053000 - Coutinho
        5203 - 1:24:12.343000 - Busquets
        5211 - 1:32:37.320000 - Ramos
        """

        # Didn't play
        player_malcon = home_team.get_player_by_id(3109)
        assert player_malcon not in minutes_played_map

        # Started second half
        player_coutinho = home_team.get_player_by_id(3501)
        assert minutes_played_map[player_coutinho] == timedelta(
            seconds=2852.053
        )

        # Replaced in second half
        player_busquets = home_team.get_player_by_id(5203)
        assert minutes_played_map[player_busquets] == timedelta(
            seconds=5052.343
        )

        # Played entire match
        player_ramos = home_team.get_player_by_id(5211)
        assert minutes_played_map[player_ramos] == (
            dataset.metadata.periods[0].duration
            + dataset.metadata.periods[1].duration
        )

        # Check if total difference between start and end time equal minutes played
        for item in minutes_played:
            assert item.duration.total_seconds() == pytest.approx(
                (
                    (
                        item.end_time.timestamp.total_seconds()
                        + __period_offset(
                            period_id=item.end_time.period.id, dataset=dataset
                        )
                    )
                    - (
                        item.start_time.timestamp.total_seconds()
                        + __period_offset(
                            period_id=item.start_time.period.id,
                            dataset=dataset,
                        )
                    )
                ),
                0.001,
            )


class TestAbsTimeContainer:
    def test_value_at(self, periods):
        period1, *_ = periods

        time1 = Time(period=period1, timestamp=timedelta(seconds=800))
        container = TimeContainer()
        container[time1] = 10

        value = container.value_at(time1 + timedelta(seconds=1))
        assert value == 10

        value = container.value_at(time1 + timedelta(seconds=10000))
        assert value == 10

        with pytest.raises(KeyError):
            container.value_at(time1 - timedelta(seconds=1))

        assert repr(container) == "TimeContainer[int]({'P1T13:20': 10})"

    def test_ranges(self, periods):
        period1, period2, _ = periods

        container = TimeContainer()

        # Player gets on the pitch
        substitution_time = Time(
            period=period1, timestamp=timedelta(seconds=15 * 60)
        )
        container.set(substitution_time, "LB")

        # Switches from LB to RB
        container.set(substitution_time + timedelta(seconds=40 * 60), "RB")

        # Player gets of the pitch
        container.set(
            Time(period=period2, timestamp=timedelta(seconds=20 * 60)), None
        )

        for start, end, position in container.ranges():
            print(f"{start} - {end} = {end - start} -> {position}")

        assert container.last() is None



from dataclasses import dataclass, field
from datetime import datetime, timedelta
from typing import (
    Generic,
    List,
    Literal,
    Optional,
    Tuple,
    TypeVar,
    Union,
    overload,
)

from sortedcontainers import SortedDict

from kloppy.exceptions import KloppyError


@dataclass
class Period:
    """
    Period

    Attributes:
        id: `1` for first half, `2` for second half, `3` for first half of
            overtime, `4` for second half of overtime, `5` for penalty shootout
        start_timestamp: The UTC datetime of the kick-off or, if the
            absolute datetime is not available, the offset between the start
            of the data feed and the period's kick-off
        start_time: Same as `start_timestamp`, but as a [`Time`][kloppy.domain.Time] object.
        end_timestamp: The UTC datetime of the final whistle or, if the
            absolute datetime is not available, the offset between the start
            of the data feed and the period's final whistle
        end_time: Same as `end_timestamp`, but as a [`Time`][kloppy.domain.Time] object.
        duration: The length of the period.
        prev_period: Period before this period.
        next_period: Period after this period.
    """

    id: int
    start_timestamp: Union[datetime, timedelta]
    end_timestamp: Optional[Union[datetime, timedelta]]

    prev_period: Optional["Period"] = field(init=False)
    next_period: Optional["Period"] = field(init=False)

    def __post_init__(self):
        if (
            self.end_timestamp is not None
            and self.start_timestamp.__class__ != self.end_timestamp.__class__
        ):
            raise ValueError(
                "'start_timestamp' and 'end_timestamp' must both be of type datetime or timedelta,"
                + f" but got start_timestamp={self.start_timestamp.__class__}"
                + f" and end_timestamp={self.end_timestamp.__class__}"
            )

    def contains(self, timestamp: datetime):
        if isinstance(self.start_timestamp, datetime) and isinstance(
            self.end_timestamp, datetime
        ):
            return self.start_timestamp <= timestamp <= self.end_timestamp
        raise KloppyError(
            "This method can only be used when start_timestamp and end_timestamp are a datetime"
        )

    @property
    def start_time(self) -> "Time":
        return Time(period=self, timestamp=timedelta(0))

    @property
    def end_time(self) -> "Time":
        return Time(period=self, timestamp=self.end_timestamp - self.start_timestamp)  # type: ignore

    @property
    def duration(self) -> timedelta:
        return self.end_timestamp - self.start_timestamp  # type: ignore

    def __eq__(self, other):
        return isinstance(other, Period) and other.id == self.id

    def __lt__(self, other: "Period"):
        return self.id < other.id

    def __ge__(self, other):
        return self == other or other < self

    def __hash__(self):
        return id(self.id)

    def set_refs(
        self,
        prev: Optional["Period"],
        next_: Optional["Period"],
    ):
        """
        Set references to other periods

        Parameters:
            prev: Period before this period
            next_: Period after this period
        """
        self.prev_period = prev
        self.next_period = next_


@dataclass
class Time:
    """
    A point in time during a match.

    A time is defined by a period and a timestamp. The timestamp is the time
    elapsed since the start of the period.

    Attributes:
        period: A period in a match.
        timestamp: The time elapsed since the start of the period.
    """

    period: "Period"
    timestamp: timedelta

    @classmethod
    def from_period(
        cls,
        period: Period,
        type_: Union[Literal["start"], Literal["end"]] = "start",
    ):
        return cls(
            period=period,
            timestamp=timedelta(0) if type_ == "start" else period.duration,
        )

    @overload
    def __sub__(self, other: timedelta) -> "Time":
        ...

    @overload
    def __sub__(self, other: "Time") -> timedelta:
        ...

    def __sub__(
        self, other: Union["Time", timedelta]
    ) -> Union["Time", timedelta]:
        """
        Subtract a timedelta or AbsTime from the current AbsTime.

        AbsTime - AbsTime = timedelta
        AbsTime - timedelta = AbsTime

        The period duration must be taking into account.
        """
        if isinstance(other, timedelta):
            current_timestamp = self.timestamp
            current_period = self.period
            while other > current_timestamp:
                other -= current_timestamp
                if not current_period.prev_period:
                    # We reached start of the match, let's just return start itself
                    return Time(period=current_period, timestamp=timedelta(0))

                current_period = current_period.prev_period
                current_timestamp = current_period.duration

            return Time(
                period=current_period, timestamp=current_timestamp - other
            )

        elif isinstance(other, Time):
            if self.period >= other.period:
                diff = self.timestamp
                current_period = self.period
                while current_period > other.period:
                    current_period = current_period.prev_period
                    diff += current_period.duration

                return diff - other.timestamp
            else:
                return -other.__sub__(self)
        else:
            raise ValueError(f"Cannot subtract {other}")

    def __add__(self, other: timedelta) -> "Time":
        assert isinstance(other, timedelta)
        current_timestamp = self.timestamp
        current_period = self.period
        while (current_timestamp + other) > current_period.duration:
            # Subtract time left in this period

            other -= current_period.duration - current_timestamp
            if not current_period.next_period:
                # We reached start of the match, let's just return start itself
                return Time(
                    period=current_period, timestamp=current_period.duration
                )

            current_period = current_period.next_period
            current_timestamp = timedelta(0)

        return Time(period=current_period, timestamp=current_timestamp + other)

    def __radd__(self, other: timedelta) -> "Time":
        assert isinstance(other, timedelta)
        return self.__add__(other)

    def __rsub__(self, other):
        raise RuntimeError("Doesn't make sense.")

    def __lt__(self, other):
        return self.period < other.period or (
            self.period == other.period and self.timestamp < other.timestamp
        )

    def __str__(self):
        m, s = divmod(self.timestamp.total_seconds(), 60)
        return f"P{self.period.id}T{m:02.0f}:{s:02.0f}"

    def __hash__(self):
        return hash((self.period.id, self.timestamp.total_seconds()))


SENTINEL = object()

T = TypeVar("T")


class TimeContainer(Generic[T]):
    def __init__(self):
        self.items: SortedDict = SortedDict()

    def set(self, time: Time, item: Optional[T]):
        self.items[time] = item  # Pair(key=time, item=item)

    def value_at(self, time: Time) -> Optional[T]:
        idx = self.items.bisect_right(time) - 1
        if idx < 0:
            raise KeyError("Not found")
        return self.items.values()[idx]

    def __getitem__(self, item: Time):
        return self.value_at(item)

    def __setitem__(self, key: Time, value: Optional[T]):
        self.set(key, value)

    def ranges(self) -> List[Tuple[Time, Time, T]]:
        items = list(self.items)
        if not items:
            return []

        # When last item isn't None (meaning 'end'), make sure to add it
        if self.items[items[-1]] is not None:
            current_period = items[0].period
            while current_period.next_period:
                current_period = current_period.next_period
            items.append(current_period.end_time)

        if len(items) < 2:
            raise ValueError("Cannot create ranges when length < 2")

        ranges_ = []
        for start_time, end_time in zip(items[:-1], items[1:]):
            ranges_.append((start_time, end_time, self.items[start_time]))
        return ranges_

    def last(self, include_time: bool = False, default=SENTINEL):
        if not len(self.items):
            if default == SENTINEL:
                raise KeyError
            else:
                return default

        time = self.items.keys()[-1]
        if include_time:
            return time, self.items[time]
        else:
            return self.items[time]

    def at_start(self):
        """Return the value at the beginning of the match"""
        if not self.items:
            raise KeyError

        first_item: Time = self.items.keys()[0]

        tmp_period = first_item.period
        while tmp_period.prev_period:
            tmp_period = tmp_period.prev_period

        return self.value_at(Time.from_period(tmp_period, "start"))

    def __repr__(self):
        if not self.items:
            return "<TimeContainer>"

        item_type = type(self.items.values()[0]).__name__
        str_items = {str(key): value for key, value in self.items.items()}
        return f"TimeContainer[{item_type}]({str_items})"

    def reset(self):
        """Clears all items from the container."""
        self.items.clear()



from abc import ABC, abstractmethod
from typing import List, NamedTuple

from kloppy.domain import EventDataset


class EventDatasetAggregator(ABC):
    @abstractmethod
    def aggregate(self, dataset: EventDataset) -> List[NamedTuple]:
        raise NotImplementedError



from datetime import timedelta
from pathlib import Path

import pytest

from kloppy import statsbomb
from kloppy.domain import EventDataset, Point
from kloppy.domain.services.transformers.attribute import (
    AngleToGoalTransformer,
    DistanceToGoalTransformer,
    DistanceToOwnGoalTransformer,
)


class TestToRecords:
    @pytest.fixture
    def event_data(self, base_dir) -> str:
        return base_dir / "files/statsbomb_event.json"

    @pytest.fixture
    def lineup_data(self, base_dir) -> str:
        return base_dir / "files/statsbomb_lineup.json"

    @pytest.fixture
    def dataset(self, event_data: Path, lineup_data: Path) -> EventDataset:
        return statsbomb.load(
            lineup_data=lineup_data,
            event_data=event_data,
            coordinates="statsbomb",
        )

    def test_default_columns(self, dataset: EventDataset):
        records = dataset.to_records()
        assert len(records) == 4061
        assert list(records[0].keys()) == [
            "event_id",
            "event_type",
            "period_id",
            "timestamp",
            "end_timestamp",
            "ball_state",
            "ball_owning_team",
            "team_id",
            "player_id",
            "coordinates_x",
            "coordinates_y",
            "result",
            "success",
        ]

    def test_string_columns(self, dataset: EventDataset):
        """
        When string columns passed to to_records the data must be searched from:
        1. the output of Default()
        2. attributes of the event
        """

        records = dataset.filter("pass").to_records(
            "timestamp", "coordinates_x", "coordinates"
        )
        assert records[0] == {
            "timestamp": timedelta(seconds=0.098),
            "coordinates_x": 60.5,
            "coordinates": Point(x=60.5, y=40.5),
        }

    def test_string_wildcard_columns(self, dataset: EventDataset):
        """
        Make sure it's possible to specify wildcard pattern to match attributes.
        """

        records = dataset.filter("pass").to_records(
            "timestamp",
            "player_id",
            "coordinates_*",
            DistanceToGoalTransformer(),
            DistanceToOwnGoalTransformer(),
            AngleToGoalTransformer(),
        )
        assert records[0] == {
            "timestamp": timedelta(seconds=0.098),
            "player_id": "6581",
            "coordinates_x": 60.5,
            "coordinates_y": 40.5,
            "distance_to_goal": 59.50210080324896,
            "distance_to_own_goal": 60.502066080424065,
            "angle_to_goal": 89.49633196102769,
        }

    @pytest.mark.parametrize(
        "coordinate_system", ["statsbomb", "tracab", "opta"]
    )
    def test_angle_to_goal_transformer(
        self, event_data: Path, lineup_data: Path, coordinate_system
    ):
        """
        Make sure calculation of the angle is consistent in different coordinate systems.
        """
        dataset = statsbomb.load(
            lineup_data=lineup_data,
            event_data=event_data,
            coordinates=coordinate_system,
        )
        records = dataset.filter("pass").to_records(
            "timestamp",
            AngleToGoalTransformer(),
        )
        assert records[0] == {
            "timestamp": timedelta(seconds=0.098),
            "angle_to_goal": 89.49633196102769,
        }



from datetime import datetime, timezone
from typing import IO, List, Optional, Tuple

from lxml import objectify

from kloppy.domain import Ground, Period, Team

from .base import TracabMetadataParser
from .common import create_period, create_team


class TracabHierarchicalXMLMetadataParser(TracabMetadataParser):
    """Extract metadata from an XML file with a hierarchical metadata object.

    Expected XML structure:

        <match iId="1" ...>
            <period iId="1" iStartFrame="1848508" iEndFrame="1916408"/>
            ...
        </match>

    Args:
        feed : The data stream of a game to parse.

    Attributes:
        root : The root of the XML tree.
    """

    def __init__(self, feed: IO[bytes]) -> None:
        self.root = objectify.fromstring(feed.read())

    def extract_periods(self) -> List[Period]:
        frame_rate = self.extract_frame_rate()
        periods = []
        for period in self.root.match.iterchildren(tag="period"):
            start_frame_id = int(period.attrib["iStartFrame"])
            end_frame_id = int(period.attrib["iEndFrame"])
            period_obj = create_period(
                int(period.attrib["iId"]),
                start_frame_id,
                end_frame_id,
                frame_rate,
            )
            if period_obj:
                periods.append(period_obj)
        return periods

    def extract_date(self) -> Optional[datetime]:
        date = datetime.strptime(
            self.root.match.attrib["dtDate"], "%Y-%m-%d %H:%M:%S"
        ).replace(tzinfo=timezone.utc)
        return date

    def extract_game_id(self) -> Optional[str]:
        game_id = self.root.match.attrib["iId"]
        return game_id

    def extract_lineups(self) -> Tuple[Team, Team]:
        start_frame_id = 0
        for period in self.root.match.iterchildren(tag="period"):
            start_frame_id = int(period.attrib["iStartFrame"])
            break
        if hasattr(self.root, "HomeTeam") and hasattr(self.root, "AwayTeam"):
            home_team = create_team(
                self.root["HomeTeam"],
                Ground.HOME,
                start_frame_id,
                id_suffix="Id",
                player_item="Player",
            )
            away_team = create_team(
                self.root["AwayTeam"],
                Ground.AWAY,
                start_frame_id,
                id_suffix="Id",
                player_item="Player",
            )
        else:
            home_team = Team(team_id="home", name="home", ground=Ground.HOME)
            away_team = Team(team_id="away", name="away", ground=Ground.AWAY)

        return (home_team, away_team)

    def extract_pitch_dimensions(self) -> Tuple[float, float]:
        pitch_size_width = float(
            self.root.match.attrib["fPitchXSizeMeters"].replace(",", ".")
        )
        pitch_size_height = float(
            self.root.match.attrib["fPitchYSizeMeters"].replace(",", ".")
        )
        return pitch_size_width, pitch_size_height

    def extract_frame_rate(self) -> int:
        frame_rate = int(self.root.match.attrib["iFrameRateFps"])
        return frame_rate



import html
from datetime import timedelta
from typing import Dict, Optional

from kloppy.domain import (
    Period,
    Player,
    PositionType,
    Team,
)

position_types_mapping: Dict[str, PositionType] = {
    "G": PositionType.Goalkeeper,
    "D": PositionType.Defender,
    "M": PositionType.Midfielder,
    "A": PositionType.Attacker,
}


def create_team(
    team_data, ground, start_frame_id, id_suffix="Id", player_item="Player"
):
    """
    Create a team object from team data
    """
    team = Team(
        team_id=str(team_data[f"Team{id_suffix}"]),
        name=html.unescape(team_data["ShortName"]),
        ground=ground,
    )

    players = (
        team_data["Players"][player_item]
        if player_item
        else team_data["Players"]
    )
    team.players = [
        Player(
            player_id=str(player[f"Player{id_suffix}"]),
            team=team,
            first_name=html.unescape(player["FirstName"]),
            last_name=html.unescape(player["LastName"]),
            name=html.unescape(player["FirstName"] + " " + player["LastName"]),
            jersey_no=int(player["JerseyNo"]),
            starting=player["StartFrameCount"] == start_frame_id,
            starting_position=position_types_mapping.get(
                player.get("StartingPosition"), PositionType.Unknown
            ),
        )
        for player in players
    ]

    return team


def create_period(
    period_id: int, start_frame: int, end_frame: int, frame_rate: int
) -> Optional[Period]:
    """
    Create a period object if frames are valid.
    """
    if start_frame != 0 or end_frame != 0:
        return Period(
            id=period_id,
            start_timestamp=timedelta(seconds=start_frame / frame_rate),
            end_timestamp=timedelta(seconds=end_frame / frame_rate),
        )
    return None



import logging
import math
from datetime import datetime, timedelta
from typing import IO, Dict, List, NamedTuple, Optional

import pytz

from kloppy.domain import (
    BallState,
    BodyPart,
    BodyPartQualifier,
    CardQualifier,
    CardType,
    CounterAttackQualifier,
    DatasetFlag,
    DuelQualifier,
    DuelResult,
    DuelType,
    EventDataset,
    ExpectedGoals,
    GoalkeeperActionType,
    GoalkeeperQualifier,
    InterceptionResult,
    Metadata,
    Orientation,
    PassQualifier,
    PassResult,
    PassType,
    Period,
    Point,
    Point3D,
    PositionType,
    PostShotExpectedGoals,
    Provider,
    Qualifier,
    SetPieceQualifier,
    SetPieceType,
    ShotResult,
    TakeOnResult,
    Team,
)
from kloppy.exceptions import DeserializationError
from kloppy.infra.serializers.event.deserializer import EventDataDeserializer
from kloppy.utils import performance_logging

from .formation_mapping import formation_id_mapping, formation_position_mapping
from .parsers import OptaEvent, get_parser

logger = logging.getLogger(__name__)

EVENT_TYPE_START_PERIOD = 32
EVENT_TYPE_END_PERIOD = 30
EVENT_TYPE_DELETED_EVENT = 43
EVENT_TYPE_START_DELAY = 27
EVENT_TYPE_END_DELAY = 28
EVENT_TYPE_OFFSIDE_PROVOKED = 55

EVENT_TYPE_PASS = 1
EVENT_TYPE_OFFSIDE_PASS = 2
EVENT_TYPE_TAKE_ON = 3
EVENT_TYPE_TACKLE = 7
EVENT_TYPE_AERIAL = 44
EVENT_TYPE_50_50 = 67
EVENT_TYPE_INTERCEPTION = 8
EVENT_TYPE_CLEARANCE = 12
EVENT_TYPE_SHOT_MISS = 13
EVENT_TYPE_SHOT_POST = 14
EVENT_TYPE_SHOT_SAVED = 15
EVENT_TYPE_SHOT_GOAL = 16
EVENT_TYPE_BALL_OUT = 5
EVENT_TYPE_CORNER_AWARDED = 6
EVENT_TYPE_FOUL_COMMITTED = 4
EVENT_TYPE_CARD = 17
EVENT_TYPE_RECOVERY = 49
EVENT_TYPE_TEAM_SET_UP = 34
EVENT_TYPE_FORMATION_CHANGE = 40
EVENT_TYPE_BALL_TOUCH = 61
EVENT_TYPE_BLOCKED_PASS = 74

EVENT_TYPE_SAVE = 10
EVENT_TYPE_CLAIM = 11
EVENT_TYPE_PUNCH = 41
EVENT_TYPE_KEEPER_PICK_UP = 52
EVENT_TYPE_SMOTHER = 54
KEEPER_EVENTS = [
    EVENT_TYPE_SAVE,
    EVENT_TYPE_CLAIM,
    EVENT_TYPE_PUNCH,
    EVENT_TYPE_KEEPER_PICK_UP,
    EVENT_TYPE_SMOTHER,
]

BALL_OUT_EVENTS = [EVENT_TYPE_BALL_OUT, EVENT_TYPE_CORNER_AWARDED]
DUEL_EVENTS = [
    EVENT_TYPE_TACKLE,
    EVENT_TYPE_AERIAL,
    EVENT_TYPE_50_50,
]

BALL_OWNING_EVENTS = (
    EVENT_TYPE_PASS,
    EVENT_TYPE_OFFSIDE_PASS,
    EVENT_TYPE_TAKE_ON,
    EVENT_TYPE_SHOT_MISS,
    EVENT_TYPE_SHOT_POST,
    EVENT_TYPE_SHOT_SAVED,
    EVENT_TYPE_SHOT_GOAL,
    EVENT_TYPE_RECOVERY,
    EVENT_TYPE_BALL_TOUCH,
)
DEAD_BALL_EVENTS = [
    EVENT_TYPE_BALL_OUT,
    EVENT_TYPE_CORNER_AWARDED,
    EVENT_TYPE_SHOT_GOAL,
    EVENT_TYPE_FOUL_COMMITTED,
    EVENT_TYPE_CARD,
    EVENT_TYPE_START_DELAY,
    EVENT_TYPE_END_DELAY,
    EVENT_TYPE_OFFSIDE_PROVOKED,
]

EVENT_QUALIFIER_GOAL_KICK = 124
EVENT_QUALIFIER_FREE_KICK = 5
EVENT_QUALIFIER_THROW_IN = 107
EVENT_QUALIFIER_CORNER_KICK = 6
EVENT_QUALIFIER_PENALTY = 9
EVENT_QUALIFIER_KICK_OFF = 279
EVENT_QUALIFIER_FREE_KICK_SHOT = 26

EVENT_QUALIFIER_HEAD_PASS = 3
EVENT_QUALIFIER_HEAD = 15
EVENT_QUALIFIER_LEFT_FOOT = 72
EVENT_QUALIFIER_RIGHT_FOOT = 20
EVENT_QUALIFIER_OTHER_BODYPART = 21

EVENT_QUALIFIER_LONG_BALL = 1
EVENT_QUALIFIER_CROSS = 2
EVENT_QUALIFIER_THROUGH_BALL = 4
EVENT_QUALIFIER_CHIPPED_BALL = 155
EVENT_QUALIFIER_LAUNCH = 157
EVENT_QUALIFIER_FLICK_ON = 168
EVENT_QUALIFIER_SWITCH_OF_PLAY = 196
EVENT_QUALIFIER_SHOT_ASSIST = 210
EVENT_QUALIFIER_ASSIST_2ND = 218

EVENT_QUALIFIER_FIRST_YELLOW_CARD = 31
EVENT_QUALIFIER_SECOND_YELLOW_CARD = 32
EVENT_QUALIFIER_RED_CARD = 33

EVENT_QUALIFIER_COUNTER_ATTACK = 23

EVENT_TYPE_PLAYER_OFF = 18
EVENT_TYPE_PLAYER_ON = 19
EVENT_QUALIFIER_TEAM_FORMATION = 130
EVENT_QUALIFIER_FORMATION_PLAYER_IDS = 30
EVENT_QUALIFIER_FORMATION_PLAYER_POSITIONS = 131

EVENT_QUALIFIER_XG = 321
EVENT_QUALIFIER_POST_SHOT_XG = 322

EVENT_TYPE_CHALLENGE = 45

event_type_names = {
    1: "pass",
    2: "offside pass",
    3: "take on",
    4: "foul",
    5: "out",
    6: "corner awarded",
    7: "tackle",
    8: "interception",
    9: "turnover",
    10: "save",
    11: "claim",
    12: "clearance",
    13: "miss",
    14: "post",
    15: "attempt saved",
    16: "goal",
    17: "card",
    18: "player off",
    19: "player on",
    20: "player retired",
    21: "player returns",
    22: "player becomes goalkeeper",
    23: "goalkeeper becomes player",
    24: "condition change",
    25: "official change",
    26: "unknown26",
    27: "start delay",
    28: "end delay",
    29: "unknown29",
    30: "end",
    31: "unknown31",
    32: "start",
    33: "unknown33",
    34: "team set up",
    35: "player changed position",
    36: "player changed jersey number",
    37: "collection end",
    38: "temp_goal",
    39: "temp_attempt",
    40: "formation change",
    41: "punch",
    42: "good skill",
    43: "deleted event",
    44: "aerial",
    45: "challenge",
    46: "unknown46",
    47: "rescinded card",
    48: "unknown46",
    49: "ball recovery",
    50: "dispossessed",
    51: "error",
    52: "keeper pick-up",
    53: "cross not claimed",
    54: "smother",
    55: "offside provoked",
    56: "shield ball opp",
    57: "foul throw in",
    58: "penalty faced",
    59: "keeper sweeper",
    60: "chance missed",
    61: "ball touch",
    62: "unknown62",
    63: "temp_save",
    64: "resume",
    65: "contentious referee decision",
    66: "possession data",
    67: "50/50",
    68: "referee drop ball",
    69: "failed to block",
    70: "injury time announcement",
    71: "coach setup",
    72: "caught offside",
    73: "other ball contact",
    74: "blocked pass",
    75: "delayed start",
    76: "early end",
    77: "player off pitch",
}

position_line_mapping = {
    "Goalkeeper": PositionType.Goalkeeper,
    "Defender": PositionType.Defender,
    "Midfielder": PositionType.Midfielder,
    "Forward": PositionType.Striker,
}


def _parse_pass(
    raw_event: OptaEvent,
    next_event: OptaEvent,
    next_next_event: OptaEvent,
    team: Team,
    period: Period,
) -> Dict:
    result = (
        PassResult.COMPLETE if raw_event.outcome else PassResult.INCOMPLETE
    )
    receiver_coordinates = _get_end_coordinates(raw_event.qualifiers)
    qualifiers = _get_pass_qualifiers(
        raw_event.qualifiers
    ) + _get_event_qualifiers(raw_event.qualifiers)

    # Look for a ball receipt event based on the result and sequence of events
    ball_receipt_event = None
    if result == PassResult.COMPLETE and next_event is not None:
        if (
            next_event.contestant_id == team.team_id
            and next_event.type_id in BALL_OWNING_EVENTS
        ):
            ball_receipt_event = next_event
        elif (
            next_event.type_id == EVENT_TYPE_CHALLENGE
            and next_next_event is not None
            and next_next_event.contestant_id == team.team_id
            and next_next_event.type_id in BALL_OWNING_EVENTS
        ):
            ball_receipt_event = next_next_event
    # Assign the receiver_player and receive_timestamp
    # if a ball_receipt_event was found
    if (
        ball_receipt_event is not None
        and ball_receipt_event.player_id is not None
    ):
        receiver_player = team.get_player_by_id(ball_receipt_event.player_id)
        receive_timestamp = (
            ball_receipt_event.timestamp - period.start_timestamp
        )
    else:
        receiver_player = None
        receive_timestamp = None

    # Set the end location of a deflected pass to the start location
    # of the next action and the outcome to "success" if the deflected
    # pass reached a teammate
    if (
        next_event is not None
        and next_next_event is not None
        and next_event.type_id == EVENT_TYPE_BALL_TOUCH
        and next_event.outcome == 1
        and next_next_event.contestant_id == raw_event.contestant_id
    ):
        result = PassResult.COMPLETE
        receiver_coordinates = Point(
            x=next_next_event.x,
            y=next_next_event.y,
        )
        receive_timestamp = next_next_event.timestamp - period.start_timestamp

    return dict(
        result=result,
        receiver_coordinates=receiver_coordinates,
        receiver_player=receiver_player,
        receive_timestamp=receive_timestamp,
        qualifiers=qualifiers,
    )


def _parse_offside_pass(raw_event: OptaEvent) -> Dict:
    pass_qualifiers = _get_pass_qualifiers(raw_event.qualifiers)
    overall_qualifiers = _get_event_qualifiers(raw_event.qualifiers)

    qualifiers = pass_qualifiers + overall_qualifiers

    return dict(
        result=PassResult.OFFSIDE,
        receiver_coordinates=_get_end_coordinates(raw_event.qualifiers),
        receiver_player=None,
        receive_timestamp=None,
        qualifiers=qualifiers,
    )


def _parse_take_on(raw_event: OptaEvent) -> Dict:
    if raw_event.outcome == 1:
        result = TakeOnResult.COMPLETE
    else:
        result = TakeOnResult.INCOMPLETE
    return dict(result=result)


def _parse_clearance(raw_event: OptaEvent) -> Dict:
    return dict(qualifiers=_get_event_qualifiers(raw_event.qualifiers))


def _parse_card(raw_event: OptaEvent) -> Dict:
    qualifiers = _get_event_qualifiers(raw_event.qualifiers)

    card_type = None
    for q in qualifiers:
        if isinstance(q, CardQualifier):
            card_type = q.value
            break

    return dict(result=None, qualifiers=None, card_type=card_type)


def _parse_lineup_qualifiers(raw_event: OptaEvent):
    formation_id = int(raw_event.qualifiers[EVENT_QUALIFIER_TEAM_FORMATION])
    formation = formation_id_mapping[formation_id]

    player_ids = raw_event.qualifiers[
        EVENT_QUALIFIER_FORMATION_PLAYER_IDS
    ].split(", ")
    position_ids = raw_event.qualifiers[
        EVENT_QUALIFIER_FORMATION_PLAYER_POSITIONS
    ].split(", ")

    positions_mapping = formation_position_mapping[formation]

    return formation, player_ids, position_ids, positions_mapping


def _parse_formation_change(raw_event: OptaEvent, team: Team) -> Dict:
    (
        formation,
        player_ids,
        position_ids,
        positions_mapping,
    ) = _parse_lineup_qualifiers(raw_event)

    player_positions = {}
    for player_id, position_id in zip(player_ids, position_ids):
        player = team.get_player_by_id(player_id)
        position = positions_mapping[int(position_id)]
        player_positions[player] = position

    return dict(formation_type=formation, player_positions=player_positions)


def _parse_substitution(next_event: OptaEvent, team: Team) -> Dict:
    replacement_player = None
    position = None
    if next_event.type_id == EVENT_TYPE_PLAYER_ON:
        replacement_player = team.get_player_by_id(next_event.player_id)

    raw_position_line = next_event.qualifiers.get(44)
    if raw_position_line:
        position = position_line_mapping.get(
            raw_position_line, PositionType.Unknown
        )

    return dict(replacement_player=replacement_player, position=position)


def _parse_shot(raw_event: OptaEvent) -> Dict:
    coordinates = Point(x=raw_event.x, y=raw_event.y)
    if raw_event.type_id == EVENT_TYPE_SHOT_GOAL:
        if 28 in raw_event.qualifiers:
            coordinates = Point(x=100 - raw_event.x, y=100 - raw_event.y)
            result = ShotResult.OWN_GOAL
            # ball_owning_team =
            # timestamp =
        else:
            result = ShotResult.GOAL
    elif 82 in raw_event.qualifiers:
        result = ShotResult.BLOCKED
    elif raw_event.type_id == EVENT_TYPE_SHOT_MISS:
        result = ShotResult.OFF_TARGET
    elif raw_event.type_id == EVENT_TYPE_SHOT_POST:
        result = ShotResult.OFF_TARGET
    elif raw_event.type_id == EVENT_TYPE_SHOT_SAVED:
        result = ShotResult.SAVED
    else:
        result = None

    qualifiers = _get_event_qualifiers(raw_event.qualifiers)
    result_coordinates = _get_end_coordinates(
        raw_event.qualifiers, start_coordinates=coordinates
    )
    if result == ShotResult.OWN_GOAL:
        if isinstance(result_coordinates, Point3D):
            result_coordinates = Point3D(
                x=100 - result_coordinates.x,
                y=100 - result_coordinates.y,
                z=result_coordinates.z,
            )
        elif isinstance(result_coordinates, Point):
            result_coordinates = Point(
                x=100 - result_coordinates.x,
                y=100 - result_coordinates.y,
            )

    statistics = []
    if 321 in raw_event.qualifiers:
        xg_value = float(raw_event.qualifiers[321])
        statistics.append(ExpectedGoals(value=xg_value))

    event_info = dict(
        coordinates=coordinates,
        result=result,
        result_coordinates=result_coordinates,
        qualifiers=qualifiers,
        statistics=statistics,
    )

    statistics = []
    for event_qualifier, statistic in zip(
        [EVENT_QUALIFIER_XG, EVENT_QUALIFIER_POST_SHOT_XG],
        [ExpectedGoals, PostShotExpectedGoals],
    ):
        xg_value = raw_event.qualifiers.get(event_qualifier)
        if xg_value:
            statistics.append(statistic(value=float(xg_value)))

    if statistics:
        event_info["statistics"] = statistics

    return event_info


def _parse_goalkeeper_events(raw_event: OptaEvent) -> Dict:
    qualifiers = _get_event_qualifiers(raw_event.qualifiers)
    goalkeeper_qualifiers = _get_goalkeeper_qualifiers(raw_event.type_id)
    qualifiers.extend(goalkeeper_qualifiers)

    return dict(result=None, qualifiers=qualifiers)


def _parse_duel(raw_event: OptaEvent) -> Dict:
    qualifiers = _get_event_qualifiers(raw_event.qualifiers)
    if raw_event.type_id == EVENT_TYPE_TACKLE:
        qualifiers.extend([DuelQualifier(value=DuelType.GROUND)])
    elif raw_event.type_id == EVENT_TYPE_AERIAL:
        qualifiers.extend(
            [
                DuelQualifier(value=DuelType.LOOSE_BALL),
                DuelQualifier(value=DuelType.AERIAL),
            ]
        )
    elif raw_event.type_id == EVENT_TYPE_50_50:
        qualifiers.extend(
            [
                DuelQualifier(value=DuelType.LOOSE_BALL),
                DuelQualifier(value=DuelType.GROUND),
            ]
        )

    result = DuelResult.WON if raw_event.outcome == 1 else DuelResult.LOST

    return dict(
        result=result,
        qualifiers=qualifiers,
    )


def _parse_interception(
    raw_event: OptaEvent, team: Team, next_event: OptaEvent
) -> Dict:
    qualifiers = _get_event_qualifiers(raw_event.qualifiers)
    result = InterceptionResult.SUCCESS

    if next_event is not None:
        next_event_type_id = int(next_event.type_id)
        if next_event_type_id in BALL_OUT_EVENTS:
            result = InterceptionResult.OUT
        elif (next_event_type_id in BALL_OWNING_EVENTS) and (
            next_event.contestant_id != team.team_id
        ):
            result = InterceptionResult.LOST

    return dict(
        result=result,
        qualifiers=qualifiers,
    )


def _get_end_coordinates(
    raw_qualifiers: Dict[int, str], start_coordinates: Optional[Point] = None
) -> Optional[Point]:
    x, y, z = None, None, None

    # pass
    if 140 in raw_qualifiers and 141 in raw_qualifiers:
        x = float(raw_qualifiers[140])
        y = float(raw_qualifiers[141])

    # blocked shot
    elif 146 in raw_qualifiers and 147 in raw_qualifiers:
        x = float(raw_qualifiers[146])
        y = float(raw_qualifiers[147])
        if 102 in raw_qualifiers and 103 in raw_qualifiers:
            # the goal mouth z-coordinate is projected back to the location
            # where the shot was blocked
            assert start_coordinates is not None
            x0, y0 = start_coordinates.x, start_coordinates.y
            x_proj = float(100)
            y_proj = float(raw_qualifiers[102])
            z_proj = float(raw_qualifiers[103])
            adj_proj = math.sqrt((x_proj - x0) ** 2 + (y_proj - y0) ** 2)
            adj_block = math.sqrt((x - x0) ** 2 + (y - y0) ** 2)
            z = z_proj / adj_proj * adj_block

    # passed the goal line
    elif 102 in raw_qualifiers:
        x = float(100)
        y = float(raw_qualifiers[102])
        if 103 in raw_qualifiers:
            z = float(raw_qualifiers[103])

    if x is not None and y is not None and z is not None:
        return Point3D(x=x, y=y, z=z)
    if x is not None and y is not None:
        return Point(x=x, y=y)

    return None


def _get_event_qualifiers(raw_qualifiers: Dict[int, str]) -> List[Qualifier]:
    qualifiers = []
    qualifiers.extend(_get_event_setpiece_qualifiers(raw_qualifiers))
    qualifiers.extend(_get_event_bodypart_qualifiers(raw_qualifiers))
    qualifiers.extend(_get_event_card_qualifiers(raw_qualifiers))
    qualifiers.extend(_get_event_counter_attack_qualifiers(raw_qualifiers))
    return qualifiers


def _get_pass_qualifiers(raw_qualifiers: Dict[int, str]) -> List[Qualifier]:
    qualifiers = []
    pass_qualifier_mapping = {
        EVENT_QUALIFIER_CROSS: PassType.CROSS,
        EVENT_QUALIFIER_LONG_BALL: PassType.LONG_BALL,
        EVENT_QUALIFIER_CHIPPED_BALL: PassType.CHIPPED_PASS,
        EVENT_QUALIFIER_THROUGH_BALL: PassType.THROUGH_BALL,
        EVENT_QUALIFIER_LAUNCH: PassType.LAUNCH,
        EVENT_QUALIFIER_FLICK_ON: PassType.FLICK_ON,
        EVENT_QUALIFIER_ASSIST_2ND: PassType.ASSIST_2ND,
    }
    for (
        sp_pass_qualifier,
        pass_qualifier_value,
    ) in pass_qualifier_mapping.items():
        if sp_pass_qualifier in raw_qualifiers:
            qualifiers.append(PassQualifier(value=pass_qualifier_value))

    if EVENT_QUALIFIER_SHOT_ASSIST in raw_qualifiers:
        # Qualifier '210' defines that the pass was an assist for a shot
        qualifiers.append(PassQualifier(value=PassType.SHOT_ASSIST))
        # The qualifier value then dictates whether it was a goal assist
        # or just key pass (note: values only available since May 2021).
        shot_result_qualifier = raw_qualifiers[EVENT_QUALIFIER_SHOT_ASSIST]
        if shot_result_qualifier is not None:
            if int(shot_result_qualifier) == EVENT_TYPE_SHOT_GOAL:
                qualifiers.append(PassQualifier(value=PassType.ASSIST))

    return qualifiers


def _get_event_setpiece_qualifiers(
    raw_qualifiers: Dict[int, str]
) -> List[Qualifier]:
    qualifiers = []
    if EVENT_QUALIFIER_CORNER_KICK in raw_qualifiers:
        qualifiers.append(SetPieceQualifier(value=SetPieceType.CORNER_KICK))
    elif (
        EVENT_QUALIFIER_FREE_KICK in raw_qualifiers
        or EVENT_QUALIFIER_FREE_KICK_SHOT in raw_qualifiers
    ):
        qualifiers.append(SetPieceQualifier(value=SetPieceType.FREE_KICK))
    elif EVENT_QUALIFIER_PENALTY in raw_qualifiers:
        qualifiers.append(SetPieceQualifier(value=SetPieceType.PENALTY))
    elif EVENT_QUALIFIER_THROW_IN in raw_qualifiers:
        qualifiers.append(SetPieceQualifier(value=SetPieceType.THROW_IN))
    elif EVENT_QUALIFIER_KICK_OFF in raw_qualifiers:
        qualifiers.append(SetPieceQualifier(value=SetPieceType.KICK_OFF))
    elif EVENT_QUALIFIER_GOAL_KICK in raw_qualifiers:
        qualifiers.append(SetPieceQualifier(value=SetPieceType.GOAL_KICK))
    return qualifiers


def _get_event_bodypart_qualifiers(
    raw_qualifiers: Dict[int, str]
) -> List[Qualifier]:
    qualifiers = []
    if EVENT_QUALIFIER_HEAD_PASS in raw_qualifiers:
        qualifiers.append(BodyPartQualifier(value=BodyPart.HEAD))
    elif EVENT_QUALIFIER_HEAD in raw_qualifiers:
        qualifiers.append(BodyPartQualifier(value=BodyPart.HEAD))
    elif EVENT_QUALIFIER_FLICK_ON in raw_qualifiers:
        qualifiers.append(BodyPartQualifier(value=BodyPart.HEAD))
    elif EVENT_QUALIFIER_LEFT_FOOT in raw_qualifiers:
        qualifiers.append(BodyPartQualifier(value=BodyPart.LEFT_FOOT))
    elif EVENT_QUALIFIER_RIGHT_FOOT in raw_qualifiers:
        qualifiers.append(BodyPartQualifier(value=BodyPart.RIGHT_FOOT))
    elif EVENT_QUALIFIER_OTHER_BODYPART in raw_qualifiers:
        qualifiers.append(BodyPartQualifier(value=BodyPart.OTHER))
    return qualifiers


def _get_event_card_qualifiers(
    raw_qualifiers: Dict[int, str]
) -> List[Qualifier]:
    qualifiers = []
    if EVENT_QUALIFIER_RED_CARD in raw_qualifiers:
        qualifiers.append(CardQualifier(value=CardType.RED))
    elif EVENT_QUALIFIER_FIRST_YELLOW_CARD in raw_qualifiers:
        qualifiers.append(CardQualifier(value=CardType.FIRST_YELLOW))
    elif EVENT_QUALIFIER_SECOND_YELLOW_CARD in raw_qualifiers:
        qualifiers.append(CardQualifier(value=CardType.SECOND_YELLOW))

    return qualifiers


def _get_goalkeeper_qualifiers(type_id: int) -> List[Qualifier]:
    qualifiers = []
    goalkeeper_qualifier = None
    if type_id == EVENT_TYPE_SAVE:
        goalkeeper_qualifier = GoalkeeperActionType.SAVE
    elif type_id == EVENT_TYPE_CLAIM:
        goalkeeper_qualifier = GoalkeeperActionType.CLAIM
    elif type_id == EVENT_TYPE_PUNCH:
        goalkeeper_qualifier = GoalkeeperActionType.PUNCH
    elif type_id == EVENT_TYPE_KEEPER_PICK_UP:
        goalkeeper_qualifier = GoalkeeperActionType.PICK_UP
    elif type_id == EVENT_TYPE_SMOTHER:
        goalkeeper_qualifier = GoalkeeperActionType.SMOTHER

    if goalkeeper_qualifier:
        qualifiers.append(GoalkeeperQualifier(value=goalkeeper_qualifier))

    return qualifiers


def _get_event_counter_attack_qualifiers(
    raw_qualifiers: Dict[int, str],
) -> List[Qualifier]:
    qualifiers = []
    if EVENT_QUALIFIER_COUNTER_ATTACK in raw_qualifiers:
        qualifiers.append(CounterAttackQualifier(True))

    return qualifiers


def _get_event_type_name(type_id: int) -> str:
    return event_type_names.get(type_id, "unknown")


class StatsPerformInputs(NamedTuple):
    meta_data: IO[bytes]
    meta_feed: str
    event_data: IO[bytes]
    event_feed: str
    meta_datatype: Optional[str] = None
    event_datatype: Optional[str] = None
    pitch_length: Optional[float] = None
    pitch_width: Optional[float] = None


class StatsPerformDeserializer(EventDataDeserializer[StatsPerformInputs]):
    @property
    def provider(self) -> Provider:
        return Provider.OPTA

    def deserialize(self, inputs: StatsPerformInputs) -> EventDataset:
        transformer = self.get_transformer(
            pitch_length=inputs.pitch_length, pitch_width=inputs.pitch_width
        )

        with performance_logging("load data", logger=logger):
            metadata_parser = get_parser(
                inputs.meta_data, inputs.meta_feed, inputs.event_datatype
            )
            events_parser = get_parser(
                inputs.event_data, inputs.event_feed, inputs.event_datatype
            )

        with performance_logging("parse data", logger=logger):
            periods = metadata_parser.extract_periods()
            score = metadata_parser.extract_score()
            teams = metadata_parser.extract_lineups()
            date = metadata_parser.extract_date()
            game_week = metadata_parser.extract_game_week()
            game_id = metadata_parser.extract_game_id()
            raw_events = [
                event
                for event in events_parser.extract_events()
                if event.type_id != EVENT_TYPE_DELETED_EVENT
            ]

            possession_team = None
            events = []
            for idx, raw_event in enumerate(raw_events):
                if raw_event.contestant_id == teams[0].team_id:
                    team = teams[0]
                elif raw_event.contestant_id == teams[1].team_id:
                    team = teams[1]
                else:
                    raise DeserializationError(
                        f"Unknown team_id {raw_event.contestant_id}"
                    )

                next_event = (
                    raw_events[idx + 1]
                    if (idx + 1) < len(raw_events)
                    else None
                )
                next_next_event = (
                    raw_events[idx + 2]
                    if (idx + 2) < len(raw_events)
                    else None
                )
                period = next(
                    (
                        period
                        for period in periods
                        if period.id == raw_event.period_id
                    ),
                    None,
                )
                if period is None:
                    logger.debug(
                        f"Skipping event {raw_event.id} because period doesn't match {raw_event.period_id}"
                    )
                    continue

                if raw_event.type_id == EVENT_TYPE_START_PERIOD:
                    logger.debug(
                        f"Set start of period {period.id} to {raw_event.timestamp}"
                    )
                    period.start_timestamp = raw_event.timestamp
                elif raw_event.type_id == EVENT_TYPE_END_PERIOD:
                    logger.debug(
                        f"Set end of period {period.id} to {raw_event.timestamp}"
                    )
                    period.end_timestamp = raw_event.timestamp
                elif raw_event.type_id == EVENT_TYPE_PLAYER_ON:
                    continue
                else:
                    if not period.start_timestamp:
                        # not started yet
                        continue

                    if raw_event.contestant_id == teams[0].team_id:
                        team = teams[0]
                    elif raw_event.contestant_id == teams[1].team_id:
                        team = teams[1]
                    else:
                        raise DeserializationError(
                            f"Unknown team_id {raw_event.contestant_id}"
                        )

                    player = None
                    if raw_event.player_id is not None:
                        player = team.get_player_by_id(raw_event.player_id)

                    if raw_event.type_id in BALL_OWNING_EVENTS:
                        possession_team = team

                    if raw_event.type_id in DEAD_BALL_EVENTS:
                        ball_state = BallState.DEAD
                    else:
                        ball_state = BallState.ALIVE

                    generic_event_kwargs = dict(
                        # from DataRecord
                        period=period,
                        timestamp=raw_event.timestamp - period.start_timestamp,
                        ball_owning_team=possession_team,
                        ball_state=ball_state,
                        # from Event
                        event_id=str(raw_event.id),
                        team=team,
                        player=player,
                        coordinates=Point(x=raw_event.x, y=raw_event.y),
                        raw_event=raw_event,
                    )

                    if raw_event.type_id == EVENT_TYPE_PASS:
                        pass_event_kwargs = _parse_pass(
                            raw_event,
                            next_event,
                            next_next_event,
                            team,
                            period,
                        )
                        event = self.event_factory.build_pass(
                            **pass_event_kwargs,
                            **generic_event_kwargs,
                        )
                    elif raw_event.type_id == EVENT_TYPE_OFFSIDE_PASS:
                        pass_event_kwargs = _parse_offside_pass(raw_event)
                        event = self.event_factory.build_pass(
                            **pass_event_kwargs,
                            **generic_event_kwargs,
                        )
                    elif raw_event.type_id == EVENT_TYPE_TAKE_ON:
                        take_on_event_kwargs = _parse_take_on(raw_event)
                        event = self.event_factory.build_take_on(
                            **take_on_event_kwargs,
                            **generic_event_kwargs,
                            qualifiers=None,
                        )
                    elif raw_event.type_id in (
                        EVENT_TYPE_SHOT_MISS,
                        EVENT_TYPE_SHOT_POST,
                        EVENT_TYPE_SHOT_SAVED,
                        EVENT_TYPE_SHOT_GOAL,
                    ):
                        if raw_event.type_id == EVENT_TYPE_SHOT_GOAL:
                            if 374 in raw_event.qualifiers:
                                # Qualifier 374 specifies the actual time of the shot for all goal events
                                # It uses London timezone for both MA3 and F24 feeds
                                naive_datetime = datetime.strptime(
                                    raw_event.qualifiers[374],
                                    "%Y-%m-%d %H:%M:%S.%f",
                                )
                                timezone = pytz.timezone("Europe/London")
                                aware_datetime = timezone.localize(
                                    naive_datetime
                                )
                                generic_event_kwargs["timestamp"] = (
                                    aware_datetime.astimezone(pytz.utc)
                                    - period.start_timestamp
                                )
                        shot_event_kwargs = _parse_shot(raw_event)
                        kwargs = {}
                        kwargs.update(generic_event_kwargs)
                        kwargs.update(shot_event_kwargs)
                        event = self.event_factory.build_shot(**kwargs)
                    elif raw_event.type_id == EVENT_TYPE_RECOVERY:
                        event = self.event_factory.build_recovery(
                            result=None,
                            qualifiers=None,
                            **generic_event_kwargs,
                        )
                    elif raw_event.type_id == EVENT_TYPE_CLEARANCE:
                        clearance_event_kwargs = _parse_clearance(raw_event)
                        event = self.event_factory.build_clearance(
                            result=None,
                            **clearance_event_kwargs,
                            **generic_event_kwargs,
                        )
                    elif raw_event.type_id in DUEL_EVENTS:
                        duel_event_kwargs = _parse_duel(raw_event)
                        event = self.event_factory.build_duel(
                            **duel_event_kwargs,
                            **generic_event_kwargs,
                        )
                    elif raw_event.type_id in (
                        EVENT_TYPE_INTERCEPTION,
                        EVENT_TYPE_BLOCKED_PASS,
                    ):
                        interception_event_kwargs = _parse_interception(
                            raw_event, team, next_event
                        )
                        event = self.event_factory.build_interception(
                            **interception_event_kwargs,
                            **generic_event_kwargs,
                        )
                    elif raw_event.type_id in KEEPER_EVENTS:
                        # Qualifier 94 means the "save" event is a shot block by a defender
                        if 94 in raw_event.qualifiers:
                            event = self.event_factory.build_generic(
                                **generic_event_kwargs,
                                result=None,
                                qualifiers=None,
                                event_name="block",
                            )
                        else:
                            goalkeeper_event_kwargs = _parse_goalkeeper_events(
                                raw_event
                            )
                            event = self.event_factory.build_goalkeeper_event(
                                **goalkeeper_event_kwargs,
                                **generic_event_kwargs,
                            )
                    elif (raw_event.type_id == EVENT_TYPE_BALL_TOUCH) & (
                        raw_event.outcome == 0
                    ):
                        event = self.event_factory.build_miscontrol(
                            result=None,
                            qualifiers=None,
                            **generic_event_kwargs,
                        )
                    elif (raw_event.type_id == EVENT_TYPE_FOUL_COMMITTED) and (
                        raw_event.outcome == 0
                    ):
                        event = self.event_factory.build_foul_committed(
                            result=None,
                            qualifiers=None,
                            **generic_event_kwargs,
                        )
                    elif raw_event.type_id in BALL_OUT_EVENTS:
                        event = self.event_factory.build_ball_out(
                            result=None,
                            qualifiers=None,
                            **generic_event_kwargs,
                        )
                    elif raw_event.type_id == EVENT_TYPE_FORMATION_CHANGE:
                        generic_event_kwargs["timestamp"] = max(
                            timedelta(0), generic_event_kwargs["timestamp"]
                        )
                        formation_change_event_kwargs = (
                            _parse_formation_change(raw_event, team)
                        )
                        event = self.event_factory.build_formation_change(
                            result=None,
                            qualifiers=None,
                            **formation_change_event_kwargs,
                            **generic_event_kwargs,
                        )
                    elif raw_event.type_id == EVENT_TYPE_PLAYER_OFF:
                        generic_event_kwargs["timestamp"] = max(
                            timedelta(0), generic_event_kwargs["timestamp"]
                        )
                        substitution_event_kwargs = _parse_substitution(
                            next_event, team
                        )
                        event = self.event_factory.build_substitution(
                            result=None,
                            qualifiers=None,
                            **substitution_event_kwargs,
                            **generic_event_kwargs,
                        )

                    elif raw_event.type_id == EVENT_TYPE_CARD:
                        card_event_kwargs = _parse_card(raw_event)

                        event = self.event_factory.build_card(
                            **card_event_kwargs,
                            **generic_event_kwargs,
                        )
                    else:
                        event = self.event_factory.build_generic(
                            **generic_event_kwargs,
                            result=None,
                            qualifiers=None,
                            event_name=_get_event_type_name(raw_event.type_id),
                        )

                    if self.should_include_event(event):
                        events.append(transformer.transform_event(event))

        metadata = Metadata(
            teams=list(teams),
            periods=periods,
            pitch_dimensions=transformer.get_to_coordinate_system().pitch_dimensions,
            score=score,
            frame_rate=None,
            orientation=Orientation.ACTION_EXECUTING_TEAM,
            flags=DatasetFlag.BALL_OWNING_TEAM | DatasetFlag.BALL_STATE,
            provider=(
                Provider.OPTA
                if inputs.event_feed.upper() == "F24"
                else Provider.STATSPERFORM
            ),
            coordinate_system=transformer.get_to_coordinate_system(),
            date=date,
            game_week=game_week,
            game_id=game_id,
        )

        return EventDataset(
            metadata=metadata,
            records=events,
        )



from datetime import datetime, timedelta, timezone

import pytest

from kloppy import datafactory
from kloppy.domain import (
    DatasetType,
    Ground,
    Orientation,
    Point,
    Provider,
    SetPieceType,
)


class TestDatafactory:
    @pytest.fixture
    def event_data(self, base_dir) -> str:
        return base_dir / "files/datafactory_events.json"

    def test_correct_deserialization(self, event_data: str):
        dataset = datafactory.load(
            event_data=event_data, coordinates="datafactory"
        )

        assert dataset.metadata.provider == Provider.DATAFACTORY
        assert dataset.dataset_type == DatasetType.EVENT
        assert len(dataset.events) == 1027
        assert len(dataset.metadata.periods) == 2
        assert dataset.events[10].ball_owning_team == dataset.metadata.teams[1]
        assert dataset.events[23].ball_owning_team == dataset.metadata.teams[0]
        assert dataset.metadata.orientation == Orientation.HOME_AWAY
        assert dataset.metadata.teams[0].name == "Team A"
        assert dataset.metadata.teams[0].ground == Ground.HOME
        assert dataset.metadata.teams[1].name == "Team B"
        assert dataset.metadata.teams[1].ground == Ground.AWAY

        player = dataset.metadata.teams[0].players[0]
        assert player.player_id == "38804"
        assert player.jersey_no == 1
        assert str(player) == "Daniel Bold"
        assert player.starting_position is None
        assert player.starting

        assert dataset.metadata.periods[0].id == 1
        assert dataset.metadata.periods[0].start_timestamp == datetime(
            2011, 11, 11, 9, 0, 13, 0, timezone.utc
        )
        assert dataset.metadata.periods[0].end_timestamp == datetime(
            2011, 11, 11, 9, 48, 45, 0, timezone.utc
        )
        assert dataset.metadata.periods[1].id == 2
        assert dataset.metadata.periods[1].start_timestamp == datetime(
            2011, 11, 11, 10, 3, 45, 0, timezone.utc
        )
        assert dataset.metadata.periods[1].end_timestamp == datetime(
            2011, 11, 11, 10, 53, 55, 0, timezone.utc
        )

        # Check enriched metadata
        date = dataset.metadata.date
        if date:
            assert isinstance(date, datetime)
            assert date == datetime(2011, 11, 11, 0, 0, tzinfo=timezone.utc)

        game_week = dataset.metadata.game_week
        if game_week:
            assert isinstance(game_week, str)
            assert game_week == "Final"

        game_id = dataset.metadata.game_id
        if game_id:
            assert isinstance(game_id, str)
            assert game_id == "1111111"

        assert dataset.events[0].timestamp == timedelta(
            seconds=3
        )  # kickoff first half
        assert dataset.events[473].timestamp == timedelta(
            seconds=4
        )  # kickoff second half

        assert dataset.events[0].coordinates == Point(0.01, 0.01)

        # Check the qualifiers
        assert dataset.events[0].qualifiers[0].value == SetPieceType.KICK_OFF
        assert dataset.events[412].qualifiers[0].value == SetPieceType.THROW_IN

    def test_correct_normalized_deserialization(self, event_data: str):
        dataset = datafactory.load(event_data=event_data)

        assert dataset.events[0].coordinates.x == pytest.approx(0.505, 0.001)
        assert dataset.events[0].coordinates.y == pytest.approx(0.505, 0.001)



from datetime import timedelta
from typing import IO, Iterator, List, Tuple

from kloppy.domain import (
    BallState,
    Frame,
    Period,
    Player,
    PlayerData,
    Point,
    Point3D,
    Team,
)
from kloppy.domain.services.frame_factory import create_frame
from kloppy.exceptions import DeserializationError

from .base import TracabDataParser


class TracabDatParser(TracabDataParser):
    def __init__(
        self,
        feed: IO[bytes],
        periods: List[Period],
        teams: Tuple[Team, Team],
        frame_rate: int,
    ) -> None:
        self.root = feed.readlines()
        self.periods = periods
        self.teams = teams
        self.frame_rate = frame_rate

    def extract_frames(
        self, sample_rate: float, only_alive: bool
    ) -> Iterator[Frame]:
        n = 0
        sample = 1.0 / sample_rate

        for line in self.root:
            line = line.strip().decode("ascii")
            if not line:
                continue

            frame_id = int(line[:10].split(":", 1)[0])
            if only_alive and not line.endswith("Alive;:"):
                continue

            for period in self.periods:
                assert isinstance(
                    period.start_timestamp, timedelta
                ), "The period's start_timestamp should be a relative time (i.e., a timedelta object)"
                assert isinstance(
                    period.end_timestamp, timedelta
                ), "The period's start_timestamp should be a relative time (i.e., a timedelta object)"

                if (
                    period.start_timestamp
                    <= timedelta(seconds=frame_id / self.frame_rate)
                    <= period.end_timestamp
                ):
                    if n % sample == 0:
                        frame = self._parse_frame(period, line)
                        yield frame
                    n += 1

    def _parse_frame(self, period, line) -> Frame:
        line = str(line)
        frame_id, players, ball = line.strip().split(":")[:3]

        players_data = {}

        for player_data in players.split(";")[:-1]:
            team_id, target_id, jersey_no, x, y, speed = player_data.split(",")
            team_id = int(team_id)

            if team_id == 1:
                team = self.teams[0]
            elif team_id == 0:
                team = self.teams[1]
            elif team_id in (-1, 3, 4):
                continue
            else:
                raise DeserializationError(
                    f"Unknown Player Team ID: {team_id}"
                )

            player = team.get_player_by_jersey_number(jersey_no)

            if not player:
                player = Player(
                    player_id=f"{team.ground}_{jersey_no}",
                    team=team,
                    jersey_no=int(jersey_no),
                )
                team.players.append(player)

            players_data[player] = PlayerData(
                coordinates=Point(float(x), float(y)), speed=float(speed)
            )

        (
            ball_x,
            ball_y,
            ball_z,
            ball_speed,
            ball_owning_team,
            ball_state,
        ) = ball.rstrip(";").split(",")[:6]

        frame_id = int(frame_id)

        if ball_owning_team == "H":
            ball_owning_team = self.teams[0]
        elif ball_owning_team == "A":
            ball_owning_team = self.teams[1]
        else:
            raise DeserializationError(
                f"Unknown ball owning team: {ball_owning_team}"
            )

        if ball_state == "Alive":
            ball_state = BallState.ALIVE
        elif ball_state == "Dead":
            ball_state = BallState.DEAD
        else:
            raise DeserializationError(f"Unknown ball state: {ball_state}")

        frame = create_frame(
            frame_id=frame_id,
            timestamp=timedelta(seconds=frame_id / self.frame_rate)
            - period.start_timestamp,
            ball_coordinates=Point3D(
                float(ball_x), float(ball_y), float(ball_z)
            ),
            ball_state=ball_state,
            ball_owning_team=ball_owning_team,
            players_data=players_data,
            period=period,
            other_data={},
        )

        return frame



from typing import List, Optional

from kloppy.config import get_config
from kloppy.domain import EventDataset, EventFactory, TrackingDataset
from kloppy.infra.serializers.event.sportec import (
    SportecEventDataDeserializer,
    SportecEventDataInputs,
)
from kloppy.infra.serializers.tracking.sportec import (
    SportecTrackingDataDeserializer,
    SportecTrackingDataInputs,
)
from kloppy.io import FileLike, open_as_file
from kloppy.utils import deprecated


def load_event(
    event_data: FileLike,
    meta_data: FileLike,
    event_types: Optional[List[str]] = None,
    coordinates: Optional[str] = None,
    event_factory: Optional[EventFactory] = None,
) -> EventDataset:
    """
    Load Sportec Solutions event data.

    Args:
        event_data: XML feed with the raw event data of a game.
        meta_data: XML feed containing the metadata of the game.
        event_types: A list of event types to load.
        coordinates: The coordinate system to use.
        event_factory: A custom event factory.

    Returns:
        The parsed event data.
    """
    serializer = SportecEventDataDeserializer(
        event_types=event_types,
        coordinate_system=coordinates,
        event_factory=event_factory or get_config("event_factory"),
    )
    with open_as_file(event_data) as event_data_fp, open_as_file(
        meta_data
    ) as meta_data_fp:
        return serializer.deserialize(
            SportecEventDataInputs(
                event_data=event_data_fp, meta_data=meta_data_fp
            )
        )


def load_tracking(
    meta_data: FileLike,
    raw_data: FileLike,
    sample_rate: Optional[float] = None,
    limit: Optional[int] = None,
    coordinates: Optional[str] = None,
    only_alive: Optional[bool] = True,
) -> TrackingDataset:
    """
    Load Sportec Solutions tracking data.

    Args:
        meta_data: A json feed containing the metadata of the game.
        raw_data: A json feed containing the raw tracking data.
        sample_rate: Sample the data at a specific rate.
        limit: Limit the number of frames to load to the first `limit` frames.
        coordinates: The coordinate system to use.
        only_alive: Only include frames in which the game is not paused.

    Returns:
        The parsed tracking data.
    """
    deserializer = SportecTrackingDataDeserializer(
        sample_rate=sample_rate,
        limit=limit,
        coordinate_system=coordinates,
        only_alive=only_alive,
    )
    with open_as_file(meta_data) as meta_data_fp, open_as_file(
        raw_data
    ) as raw_data_fp:
        return deserializer.deserialize(
            inputs=SportecTrackingDataInputs(
                meta_data=meta_data_fp, raw_data=raw_data_fp
            )
        )


@deprecated("sportec.load_event should be used")
def load(
    event_data: FileLike,
    meta_data: FileLike,
    event_types: Optional[List[str]] = None,
    coordinates: Optional[str] = None,
    event_factory: Optional[EventFactory] = None,
) -> EventDataset:
    return load_event(
        event_data, meta_data, event_types, coordinates, event_factory
    )


def get_IDSSE_url(match_id: str, data_type: str) -> str:
    """Returns the URL for the meta, event or tracking data for a match in the IDDSE dataset."""
    # match_id -> file_id
    DATA_MAP = {
        "J03WPY": {"meta": 51643487, "event": 51643505, "tracking": 51643526},
        "J03WN1": {"meta": 51643472, "event": 51643496, "tracking": 51643517},
        "J03WMX": {"meta": 51643475, "event": 51643493, "tracking": 51643514},
        "J03WOH": {"meta": 51643478, "event": 51643499, "tracking": 51643520},
        "J03WQQ": {"meta": 51643484, "event": 51643508, "tracking": 51643529},
        "J03WOY": {"meta": 51643481, "event": 51643502, "tracking": 51643523},
        "J03WR9": {"meta": 51643490, "event": 51643511, "tracking": 51643532},
    }
    # URL constant
    DATA_URL = (
        "https://springernature.figshare.com/ndownloader/files/{file_id}"
    )

    if data_type not in ["meta", "event", "tracking"]:
        raise ValueError(
            f"Data type should be one of ['meta', 'event', 'tracking'], but got {data_type}"
        )
    if match_id not in DATA_MAP:
        raise ValueError(
            f"This match_id is not available, please select from {list(DATA_MAP.keys())}"
        )
    return DATA_URL.format(file_id=str(DATA_MAP[match_id][data_type]))


def load_open_event_data(
    match_id: str = "J03WPY",
    event_types: Optional[List[str]] = None,
    coordinates: Optional[str] = None,
    event_factory: Optional[EventFactory] = None,
) -> EventDataset:
    """
    Load event data for a game from the IDSSE dataset.

    The IDSSE dataset will be released with the publication of the *An integrated
    dataset of synchronized spatiotemporal and event data in elite soccer*
    paper [1]_ and is released under the Creative Commons Attribution 4.0
    license.

    Args:
        match_id (str, optional):
            Match-ID of one of the matches. Defaults to `'J03WPY'`. See below
            for available matches.
        event_types:
        coordinates:
        event_factory:

    Notes:
        The dataset contains seven full matches of raw event and position data
        for both teams and the ball from the German Men's Bundesliga season
        2022/23 first and second division. A detailed description of the
        dataset as well as the collection process can be found in the
        accompanying paper.

        The following matches are available::

        matches = {
            'J03WMX': 1. FC Köln vs. FC Bayern München,
            'J03WN1': VfL Bochum 1848 vs. Bayer 04 Leverkusen,
            'J03WPY': Fortuna Düsseldorf vs. 1. FC Nürnberg,
            'J03WOH': Fortuna Düsseldorf vs. SSV Jahn Regensburg,
            'J03WQQ': Fortuna Düsseldorf vs. FC St. Pauli,
            'J03WOY': Fortuna Düsseldorf vs. F.C. Hansa Rostock,
            'J03WR9': Fortuna Düsseldorf vs. 1. FC Kaiserslautern
        }

    References:
        .. [1] Bassek, M., Rein, R., Weber, H. et al. "An integrated dataset of
               spatiotemporal and event data in elite soccer." Sci Data 12, 195 (2025).
               https://doi.org/10.1038/s41597-025-04505-y
    """
    return load_event(
        event_data=get_IDSSE_url(match_id, "event"),
        meta_data=get_IDSSE_url(match_id, "meta"),
        event_types=event_types,
        coordinates=coordinates,
        event_factory=event_factory,
    )


def load_open_tracking_data(
    match_id: str = "J03WPY",
    sample_rate: Optional[float] = None,
    limit: Optional[int] = None,
    coordinates: Optional[str] = None,
    only_alive: Optional[bool] = True,
) -> TrackingDataset:
    """
    Load tracking data for a game from the IDSSE dataset.

    The IDSSE dataset will be released with the publication of the *An integrated
    dataset of synchronized spatiotemporal and event data in elite soccer*
    paper [1]_ and is released under the Creative Commons Attribution 4.0
    license.

    Args:
        match_id (str, optional):
            Match-ID of one of the matches. Defaults to `'J03WPY'`. See below
            for available matches.
        sample_rate:
        limit:
        coordinates:
        only_alive:

    Notes:
        The dataset contains seven full matches of raw event and position data
        for both teams and the ball from the German Men's Bundesliga season
        2022/23 first and second division. A detailed description of the
        dataset as well as the collection process can be found in the
        accompanying paper.

        The following matches are available::

        matches = {
            'J03WMX': 1. FC Köln vs. FC Bayern München,
            'J03WN1': VfL Bochum 1848 vs. Bayer 04 Leverkusen,
            'J03WPY': Fortuna Düsseldorf vs. 1. FC Nürnberg,
            'J03WOH': Fortuna Düsseldorf vs. SSV Jahn Regensburg,
            'J03WQQ': Fortuna Düsseldorf vs. FC St. Pauli,
            'J03WOY': Fortuna Düsseldorf vs. F.C. Hansa Rostock,
            'J03WR9': Fortuna Düsseldorf vs. 1. FC Kaiserslautern
        }

    References:
        .. [1] Bassek, M., Weber, H., Rein, R., & Memmert, D. (2024). "An integrated
               dataset of synchronized spatiotemporal and event data in elite soccer."
               In Submission.
    """
    return load_tracking(
        raw_data=get_IDSSE_url(match_id, "tracking"),
        meta_data=get_IDSSE_url(match_id, "meta"),
        sample_rate=sample_rate,
        limit=limit,
        coordinates=coordinates,
        only_alive=only_alive,
    )



from datetime import datetime, timezone
from pathlib import Path

import pytest

from kloppy import pff
from kloppy.domain import DatasetType, Orientation, Point, Point3D, Provider


class TestPFFTracking:
    @pytest.fixture
    def meta_data_home_starts_left(self, base_dir) -> str:
        return base_dir / "files" / "pff_metadata_10517.json"

    @pytest.fixture
    def rosters_meta_data_home_starts_left(self, base_dir) -> str:
        return base_dir / "files" / "pff_rosters_10517.json"

    @pytest.fixture
    def raw_data_home_starts_left(self, base_dir) -> str:
        return base_dir / "files" / "pff_10517.jsonl"

    @pytest.fixture
    def meta_data_home_starts_right(self, base_dir) -> str:
        return base_dir / "files" / "pff_metadata_3812.json"

    @pytest.fixture
    def rosters_meta_data_home_starts_right(self, base_dir) -> str:
        return base_dir / "files" / "pff_rosters_3812.json"

    @pytest.fixture
    def raw_data_home_starts_right(self, base_dir) -> str:
        return base_dir / "files" / "pff_3812.jsonl"

    def test_correct_deserialization_alive_only(
        self,
        raw_data_home_starts_left: Path,
        meta_data_home_starts_left: Path,
        rosters_meta_data_home_starts_left: Path,
    ):
        # Raw data is obtained by grabbing first and last 25 frames of each period
        dataset = pff.load_tracking(
            meta_data=meta_data_home_starts_left,
            roster_meta_data=rosters_meta_data_home_starts_left,
            raw_data=raw_data_home_starts_left,
            coordinates="pff",
            only_alive=True,
        )
        # Check game_id
        game_id = dataset.metadata.game_id
        if game_id:
            assert isinstance(game_id, int)
            assert game_id == 10517

        # Check records size
        assert len(dataset.records) == 199

        # Check periods
        assert len(dataset.metadata.periods) == 4
        assert dataset.metadata.periods[0].id == 1
        assert dataset.metadata.periods[1].id == 2
        assert dataset.metadata.periods[2].id == 3
        assert dataset.metadata.periods[3].id == 4

    def test_correct_deserialization_limit_sample(
        self,
        raw_data_home_starts_left: Path,
        meta_data_home_starts_left: Path,
        rosters_meta_data_home_starts_left: Path,
    ):

        dataset = pff.load_tracking(
            meta_data=meta_data_home_starts_left,
            roster_meta_data=rosters_meta_data_home_starts_left,
            raw_data=raw_data_home_starts_left,
            coordinates="pff",
            only_alive=True,
            limit=100,
        )
        assert len(dataset.records) == 100

        dataset = pff.load_tracking(
            meta_data=meta_data_home_starts_left,
            roster_meta_data=rosters_meta_data_home_starts_left,
            raw_data=raw_data_home_starts_left,
            coordinates="pff",
            only_alive=True,
            limit=100,
            sample_rate=(1 / 2),
        )
        assert len(dataset.records) == 100

    def test_correct_deserialization(
        self,
        raw_data_home_starts_left: Path,
        meta_data_home_starts_left: Path,
        rosters_meta_data_home_starts_left: Path,
    ):
        # Raw data is obtained by grabbing first and last 25 frames of each period
        dataset = pff.load_tracking(
            meta_data=meta_data_home_starts_left,
            roster_meta_data=rosters_meta_data_home_starts_left,
            raw_data=raw_data_home_starts_left,
            coordinates="pff",
            only_alive=False,
        )

        # Check Provider is PFF and DatasetType is Tracking
        assert dataset.metadata.provider == Provider.PFF
        assert dataset.dataset_type == DatasetType.TRACKING

        # Check game_id
        game_id = dataset.metadata.game_id
        if game_id:
            assert isinstance(game_id, int)
            assert game_id == 10517

        # Check records size
        assert len(dataset.records) == 200

        # Check periods
        assert len(dataset.metadata.periods) == 4
        assert dataset.metadata.periods[0].id == 1
        assert dataset.metadata.periods[1].id == 2
        assert dataset.metadata.periods[2].id == 3
        assert dataset.metadata.periods[3].id == 4
        # Check Pitch Dimensions
        pitch_dimensions = dataset.metadata.pitch_dimensions
        assert pitch_dimensions.x_dim.min == -52.5
        assert pitch_dimensions.x_dim.max == 52.5
        assert pitch_dimensions.y_dim.min == -34.0
        assert pitch_dimensions.y_dim.max == 34.0

        # Check Date
        date = dataset.metadata.date
        if date:
            assert isinstance(date, datetime)
            assert date == datetime(2022, 12, 18, 15, 0, tzinfo=timezone.utc)

        # Check Orientation
        assert dataset.metadata.orientation == Orientation.HOME_AWAY

        # Check first and last frame of period 1
        assert dataset.records[0].frame_id == 4630
        assert dataset.records[0].timestamp.total_seconds() == 0.000821
        assert dataset.records[49].frame_id == 99119
        assert dataset.records[49].timestamp.total_seconds() == 3152.786941

        # Check first and last frame of period 2
        assert dataset.records[50].frame_id == 100240
        assert dataset.records[50].timestamp.total_seconds() == 0.033011
        assert dataset.records[99].frame_id == 196639
        assert dataset.records[99].timestamp.total_seconds() == 3216.549528

        # Check first and last frame of period 3
        assert dataset.records[100].frame_id == 197541
        assert dataset.records[100].timestamp.total_seconds() == 0.000291
        assert dataset.records[149].frame_id == 226468
        assert dataset.records[149].timestamp.total_seconds() == 965.198823

        # Check first and last frame of period 4
        assert dataset.records[150].frame_id == 230627
        assert dataset.records[150].timestamp.total_seconds() == 0.032929
        assert dataset.records[199].frame_id == 265050
        assert dataset.records[199].timestamp.total_seconds() == 1148.614844

        # Check PFF Home Player
        assert dataset.metadata.teams[0].players[0].player_id == "13222"
        assert dataset.metadata.teams[0].players[0].name == "Nahuel Molina"
        assert dataset.metadata.teams[0].players[0].team.team_id == "364"
        assert dataset.metadata.teams[0].players[0].jersey_no == 26

        # make sure PFF ID is used as player ID for Away Player
        assert dataset.metadata.teams[1].players[0].player_id == "4622"
        assert dataset.metadata.teams[1].players[0].name == "Kingsley Coman"
        assert dataset.metadata.teams[1].players[0].team.team_id == "363"
        assert dataset.metadata.teams[1].players[0].jersey_no == 20

        # Check Home Player coordinates
        home_player = dataset.metadata.teams[0].players[4]  # Julian Alvarez
        assert dataset.records[0].players_data[
            home_player
        ].coordinates == Point(x=4.987, y=-1.993)

        # Check Away Player coordinates
        away_player = dataset.metadata.teams[1].players[17]  # Olivier Giroud
        assert dataset.records[0].players_data[
            away_player
        ].coordinates == Point(x=0.763, y=-18.099)

        # Check Ball coordinates
        assert dataset.records[0].ball_coordinates == Point3D(
            x=0.42, y=1.59, z=0.39
        )

    def test_orientation(
        self,
        raw_data_home_starts_right: Path,
        meta_data_home_starts_right: Path,
        rosters_meta_data_home_starts_right: Path,
    ):
        # Raw data is obtained by grabbing first and last 25 frames of each period
        dataset = pff.load_tracking(
            meta_data=meta_data_home_starts_right,
            roster_meta_data=rosters_meta_data_home_starts_right,
            raw_data=raw_data_home_starts_right,
            coordinates="pff",
        )

        # Check Orientation
        assert dataset.metadata.orientation == Orientation.AWAY_HOME

        # Check Player name with apostrophe (') in it
        assert dataset.metadata.teams[0].players[1].name == "Moussa N'Diaye"

    def test_correct_normalized_deserialization(
        self,
        raw_data_home_starts_left: Path,
        meta_data_home_starts_left: Path,
        rosters_meta_data_home_starts_left: Path,
    ):
        # Raw data is obtained by grabbing first and last 25 frames of each period
        dataset = pff.load_tracking(
            meta_data=meta_data_home_starts_left,
            roster_meta_data=rosters_meta_data_home_starts_left,
            raw_data=raw_data_home_starts_left,
        )

        # Check Pitch Dimensions
        pitch_dimensions = dataset.metadata.pitch_dimensions
        assert pitch_dimensions.x_dim.min == 0
        assert pitch_dimensions.x_dim.max == 1
        assert pitch_dimensions.y_dim.min == 0
        assert pitch_dimensions.y_dim.max == 1

        home_player = dataset.metadata.teams[0].players[4]
        assert dataset.records[0].players_data[
            home_player
        ].coordinates == Point(x=0.547495238095238, y=0.5293088235294118)

        # Check Away Player coordinates
        away_player = dataset.metadata.teams[1].players[17]  # Olivier Giroud
        assert dataset.records[0].players_data[
            away_player
        ].coordinates == Point(x=0.5072666666666666, y=0.7661617647058824)

        # Check Ball coordinates
        assert dataset.records[0].ball_coordinates == Point3D(
            x=0.504, y=0.4766176470588235, z=0.39
        )



"""Base class for all parsers that can handle Tracab metadata files.

A parser reads a single metadata file and should extend the 'TracabMetadataParser'
class to extract the data about periods, lineups, pitch dimensions, etc.
"""

from abc import ABC, abstractmethod
from datetime import datetime
from typing import IO, List, Optional, Tuple

from kloppy.domain import Orientation, Period, Score, Team


class TracabMetadataParser(ABC):
    """Extract data from a tracab metadata file."""

    def __init__(self, feed: IO[bytes]) -> None:
        """Initialize the parser with the data stream.

        Args:
            feed : The metadata of a game to parse.
        """

    @abstractmethod
    def extract_periods(self) -> List[Period]:
        """Extract the periods of the game."""

    def extract_score(self) -> Optional[Score]:
        """Extract the game's score."""
        return None

    def extract_date(self) -> Optional[datetime]:
        """Extract the game's date."""
        return None

    def extract_game_week(self) -> Optional[str]:
        """Extract the game week."""
        return None

    def extract_game_id(self) -> Optional[str]:
        """Extract the game's id."""
        return None

    @abstractmethod
    def extract_lineups(self) -> Tuple[Team, Team]:
        """Extract the home and away team."""

    @abstractmethod
    def extract_pitch_dimensions(self) -> Tuple[float, float]:
        """Extract the pitch size as (length, width)."""

    @abstractmethod
    def extract_frame_rate(self) -> int:
        """Extract the tracking data's frame rate."""

    def extract_orientation(self) -> Optional[Orientation]:
        """Extract the orientation of the data."""
        return None



import sys

import pytest
from pandas import DataFrame
from pandas.testing import assert_frame_equal

from kloppy import opta, statsbomb, tracab
from kloppy.config import config_context
from kloppy.domain import (
    AttackingDirection,
    DatasetFlag,
    Dimension,
    Ground,
    Metadata,
    MetricaCoordinateSystem,
    NormalizedPitchDimensions,
    Orientation,
    Period,
    Player,
    PlayerData,
    Point,
    Point3D,
    Provider,
    Team,
    TrackingDataset,
)
from kloppy.domain.services.frame_factory import create_frame


class TestHelpers:
    def _get_tracking_dataset(self):
        home_team = Team(team_id="home", name="home", ground=Ground.HOME)
        away_team = Team(team_id="away", name="away", ground=Ground.AWAY)
        teams = [home_team, away_team]

        periods = [
            Period(
                id=1,
                start_timestamp=0.0,
                end_timestamp=10.0,
            ),
            Period(
                id=2,
                start_timestamp=15.0,
                end_timestamp=25.0,
            ),
        ]
        metadata = Metadata(
            flags=(DatasetFlag.BALL_OWNING_TEAM),
            pitch_dimensions=NormalizedPitchDimensions(
                x_dim=Dimension(0, 100),
                y_dim=Dimension(-50, 50),
                pitch_length=105,
                pitch_width=68,
            ),
            orientation=Orientation.HOME_AWAY,
            frame_rate=25,
            periods=periods,
            teams=teams,
            score=None,
            provider=None,
            coordinate_system=None,
            date="2024-05-19T13:30:00",
            game_week="35",
            game_id="2374516",
        )

        tracking_data = TrackingDataset(
            metadata=metadata,
            records=[
                create_frame(
                    frame_id=1,
                    timestamp=0.1,
                    ball_owning_team=teams[0],
                    ball_state=None,
                    period=periods[0],
                    players_data={},
                    other_data=None,
                    ball_coordinates=Point3D(x=100, y=-50, z=0),
                ),
                create_frame(
                    frame_id=2,
                    timestamp=0.2,
                    ball_owning_team=teams[1],
                    ball_state=None,
                    period=periods[1],
                    players_data={
                        Player(
                            team=home_team, player_id="home_1", jersey_no=1
                        ): PlayerData(
                            coordinates=Point(x=15, y=35),
                            distance=0.03,
                            speed=10.5,
                            other_data={"extra_data": 1},
                        )
                    },
                    other_data={"extra_data": 1},
                    ball_coordinates=Point3D(x=0, y=50, z=1),
                ),
            ],
        )
        return tracking_data

    def test_transform(self):
        tracking_data = self._get_tracking_dataset()

        # orientation change AND dimension scale
        transformed_dataset = tracking_data.transform(
            to_orientation="AWAY_HOME",
            to_pitch_dimensions=NormalizedPitchDimensions(
                x_dim=Dimension(min=0, max=1),
                y_dim=Dimension(min=0, max=1),
                pitch_length=105,
                pitch_width=68,
            ),
        )

        assert transformed_dataset.frames[0].ball_coordinates == Point3D(
            x=0, y=1, z=0
        )
        assert transformed_dataset.frames[1].ball_coordinates == Point3D(
            x=1, y=0, z=1
        )
        assert (
            transformed_dataset.metadata.orientation == Orientation.AWAY_HOME
        )
        assert transformed_dataset.metadata.coordinate_system is None
        assert (
            transformed_dataset.metadata.pitch_dimensions
            == NormalizedPitchDimensions(
                x_dim=Dimension(min=0, max=1),
                y_dim=Dimension(min=0, max=1),
                pitch_length=105,
                pitch_width=68,
            )
        )

    def test_transform_to_pitch_dimensions(self):
        tracking_data = self._get_tracking_dataset()

        transformed_dataset = tracking_data.transform(
            to_pitch_dimensions=NormalizedPitchDimensions(
                x_dim=Dimension(min=0, max=1),
                y_dim=Dimension(min=0, max=1),
                pitch_length=105,
                pitch_width=68,
            ),
        )

        assert transformed_dataset.frames[0].ball_coordinates == Point3D(
            x=1, y=0, z=0
        )
        assert transformed_dataset.frames[1].ball_coordinates == Point3D(
            x=0, y=1, z=1
        )
        assert (
            transformed_dataset.metadata.pitch_dimensions
            == NormalizedPitchDimensions(
                x_dim=Dimension(min=0, max=1),
                y_dim=Dimension(min=0, max=1),
                pitch_length=105,
                pitch_width=68,
            )
        )

    def test_transform_to_orientation(self):
        to_pitch_dimensions = NormalizedPitchDimensions(
            x_dim=Dimension(min=0, max=1),
            y_dim=Dimension(min=0, max=1),
            pitch_length=105,
            pitch_width=68,
        )
        # Create a dataset with the KLOPPY pitch dimensions
        # and HOME_AWAY orientation
        original = self._get_tracking_dataset().transform(
            to_pitch_dimensions=to_pitch_dimensions
        )
        assert original.metadata.orientation == Orientation.HOME_AWAY
        assert original.frames[0].ball_coordinates == Point3D(x=1, y=0, z=0)
        assert original.frames[1].ball_coordinates == Point3D(x=0, y=1, z=1)
        # the frames should have the correct attacking direction
        assert original.frames[0].attacking_direction == AttackingDirection.LTR
        assert original.frames[1].attacking_direction == AttackingDirection.RTL

        # Transform to AWAY_HOME orientation
        transform1 = original.transform(
            to_orientation=Orientation.AWAY_HOME,
            to_pitch_dimensions=to_pitch_dimensions,
        )
        assert transform1.metadata.orientation == Orientation.AWAY_HOME
        # all coordinates should be flipped
        assert transform1.frames[0].ball_coordinates == Point3D(x=0, y=1, z=0)
        assert transform1.frames[1].ball_coordinates == Point3D(x=1, y=0, z=1)
        # the frames should have the correct attacking direction
        assert (
            transform1.frames[0].attacking_direction == AttackingDirection.RTL
        )
        assert (
            transform1.frames[1].attacking_direction == AttackingDirection.LTR
        )

        # Transform to STATIC_AWAY_HOME orientation
        transform2 = transform1.transform(
            to_orientation=Orientation.STATIC_AWAY_HOME,
            to_pitch_dimensions=to_pitch_dimensions,
        )
        assert transform2.metadata.orientation == Orientation.STATIC_AWAY_HOME
        # all coordintes in the second half should be flipped
        assert transform2.frames[0].ball_coordinates == Point3D(x=0, y=1, z=0)
        assert transform2.frames[1].ball_coordinates == Point3D(x=0, y=1, z=1)
        # the frames should have the correct attacking direction
        for frame in transform2.frames:
            assert frame.attacking_direction == AttackingDirection.RTL

        # Transform to BALL_OWNING_TEAM orientation
        transform3 = transform2.transform(
            to_orientation=Orientation.BALL_OWNING_TEAM,
            to_pitch_dimensions=to_pitch_dimensions,
        )
        assert transform3.metadata.orientation == Orientation.BALL_OWNING_TEAM
        # the coordinates of frame 1 should be flipped
        assert transform3.frames[0].ball_coordinates == Point3D(x=1, y=0, z=0)
        assert transform3.frames[1].ball_coordinates == Point3D(x=0, y=1, z=1)
        # the frames should have the correct attacking direction
        assert (
            transform3.frames[0].attacking_direction == AttackingDirection.LTR
        )
        assert (
            transform3.frames[1].attacking_direction == AttackingDirection.RTL
        )

        # Transform to ACTION_EXECUTING_TEAM orientation
        # this should be identical to BALL_OWNING_TEAM for tracking data
        transform4 = transform3.transform(
            to_orientation=Orientation.ACTION_EXECUTING_TEAM,
            to_pitch_dimensions=to_pitch_dimensions,
        )
        assert (
            transform4.metadata.orientation
            == Orientation.ACTION_EXECUTING_TEAM
        )
        assert transform4.frames[1].ball_coordinates == Point3D(x=0, y=1, z=1)
        for frame_t3, frame_t4 in zip(transform3.frames, transform4.frames):
            assert frame_t3.ball_coordinates == frame_t4.ball_coordinates
            assert frame_t3.attacking_direction == frame_t4.attacking_direction

        # Transform back to the original HOME_AWAY orientation
        transform5 = transform4.transform(
            to_orientation=Orientation.HOME_AWAY,
            to_pitch_dimensions=to_pitch_dimensions,
        )
        # we should be back at the original
        for frame1, frame2 in zip(original.frames, transform5.frames):
            assert frame1.ball_coordinates == frame2.ball_coordinates
            assert frame1.attacking_direction == frame2.attacking_direction

    def test_transform_to_coordinate_system(self, base_dir):
        dataset = tracab.load(
            meta_data=base_dir / "files/tracab_meta.xml",
            raw_data=base_dir / "files/tracab_raw.dat",
            only_alive=False,
            coordinates="tracab",
        )

        player_home_1 = dataset.metadata.teams[0].get_player_by_jersey_number(
            1
        )
        assert dataset.records[0].players_data[
            player_home_1
        ].coordinates == Point(x=5270.0, y=27.0)

        transformed_dataset = dataset.transform(
            to_coordinate_system=Provider.METRICA
        )
        transformerd_coordinate_system = MetricaCoordinateSystem(
            pitch_length=dataset.metadata.coordinate_system.pitch_length,
            pitch_width=dataset.metadata.coordinate_system.pitch_width,
        )

        assert transformed_dataset.records[0].players_data[
            player_home_1
        ].coordinates == Point(x=1.0019047619047619, y=0.49602941176470583)
        assert (
            transformed_dataset.metadata.orientation
            == dataset.metadata.orientation
        )
        assert (
            transformed_dataset.metadata.coordinate_system
            == transformerd_coordinate_system
        )
        assert (
            transformed_dataset.metadata.pitch_dimensions
            == transformerd_coordinate_system.pitch_dimensions
        )

    def test_transform_to_pitch_dimensions_with_coordinate_system(
        self, base_dir
    ):
        dataset = tracab.load(
            meta_data=base_dir / "files/tracab_meta.xml",
            raw_data=base_dir / "files/tracab_raw.dat",
            only_alive=False,
            coordinates="tracab",
        )

        player_home_1 = dataset.metadata.teams[0].get_player_by_jersey_number(
            1
        )
        assert dataset.records[0].players_data[
            player_home_1
        ].coordinates == Point(x=5270.0, y=27.0)

        transformed_dataset = dataset.transform(
            to_pitch_dimensions=NormalizedPitchDimensions(
                x_dim=Dimension(min=0, max=1),
                y_dim=Dimension(min=0, max=1),
                pitch_length=105,
                pitch_width=68,
            ),
        )
        assert transformed_dataset.records[0].players_data[
            player_home_1
        ].coordinates == Point(x=1.0019047619047619, y=0.5039705882352942)

        assert (
            transformed_dataset.metadata.pitch_dimensions
            == NormalizedPitchDimensions(
                x_dim=Dimension(min=0, max=1),
                y_dim=Dimension(min=0, max=1),
                pitch_length=105,
                pitch_width=68,
            )
        )
        assert (
            transformed_dataset.metadata.coordinate_system.pitch_dimensions
            == NormalizedPitchDimensions(
                x_dim=Dimension(min=0, max=1),
                y_dim=Dimension(min=0, max=1),
                pitch_length=105,
                pitch_width=68,
            )
        )

    def test_transform_event_data(self, base_dir):
        """Make sure event data that's in ACTION_EXECUTING orientation is
        transformed correctly"""
        dataset = statsbomb.load(
            lineup_data=base_dir / "files/statsbomb_lineup.json",
            event_data=base_dir / "files/statsbomb_event.json",
        )

        home_team, away_team = dataset.metadata.teams

        # This is a pressure event by Deportivo while Barcelona is in possession
        pressure_event = dataset.get_event_by_id(
            "6399af5c-74b8-4efe-ae19-85f331d355e8"
        )
        assert pressure_event.team == away_team
        assert pressure_event.ball_owning_team == home_team

        receipt_event = pressure_event.next()
        assert receipt_event.team == home_team
        assert receipt_event.ball_owning_team == home_team

        transformed_dataset = dataset.transform(
            to_orientation="static_home_away"
        )
        transformed_pressure_event = transformed_dataset.get_event_by_id(
            pressure_event.event_id
        )
        transformed_receipt_event = transformed_pressure_event.next()

        # The receipt event is executed by the away team and should be changed by the transformation
        assert (
            pressure_event.coordinates.x
            == 1 - transformed_pressure_event.coordinates.x
        )
        assert (
            pressure_event.coordinates.y
            == 1 - transformed_pressure_event.coordinates.y
        )

        # The receipt event is executed by the home team and shouldn't be changed by the transformation
        assert (
            receipt_event.coordinates.x
            == transformed_receipt_event.coordinates.x
        )
        assert (
            receipt_event.coordinates.y
            == transformed_receipt_event.coordinates.y
        )

    def test_transform_event_data_freeze_frame(self, base_dir):
        """Make sure the freeze frame within event data is transformed too"""
        dataset = statsbomb.load(
            lineup_data=base_dir / "files/statsbomb_lineup.json",
            event_data=base_dir / "files/statsbomb_event.json",
        )

        _, away_team = dataset.metadata.teams

        shot_event = dataset.get_event_by_id(
            "65f16e50-7c5d-4293-b2fc-d20887a772f9"
        )
        transformed_dataset = dataset.transform(
            to_orientation="static_away_home"
        )
        shot_event_transformed = transformed_dataset.get_event_by_id(
            shot_event.event_id
        )

        player = away_team.get_player_by_id(6612)
        coordinates = shot_event.freeze_frame.players_coordinates[player]
        coordinates_transformed = (
            shot_event_transformed.freeze_frame.players_coordinates[player]
        )

        assert coordinates.x == 1 - coordinates_transformed.x
        assert coordinates.y == 1 - coordinates_transformed.y

    def test_to_pandas(self):
        tracking_data = self._get_tracking_dataset()

        data_frame = tracking_data.to_df(engine="pandas")

        expected_data_frame = DataFrame.from_dict(
            {
                "frame_id": {0: 1, 1: 2},
                "period_id": {0: 1, 1: 2},
                "timestamp": {0: 0.1, 1: 0.2},
                "ball_state": {0: None, 1: None},
                "ball_owning_team_id": {0: "home", 1: "away"},
                "ball_x": {0: 100, 1: 0},
                "ball_y": {0: -50, 1: 50},
                "ball_z": {0: 0, 1: 1},
                "ball_speed": {0: None, 1: None},
                "home_1_x": {0: None, 1: 15.0},
                "home_1_y": {0: None, 1: 35.0},
                "home_1_d": {0: None, 1: 0.03},
                "home_1_s": {0: None, 1: 10.5},
                "home_1_extra_data": {0: None, 1: 1},
                "extra_data": {0: None, 1: 1},
            }
        )
        assert_frame_equal(data_frame, expected_data_frame, check_like=True)

    def test_to_pandas_generic_events(self, base_dir):
        dataset = opta.load(
            f7_data=base_dir / "files/opta_f7.xml",
            f24_data=base_dir / "files/opta_f24.xml",
        )

        dataframe = dataset.to_df(engine="pandas")
        dataframe = dataframe[dataframe.event_type == "BALL_OUT"]
        assert dataframe.shape[0] == 2

    def test_to_pandas_incomplete_pass(self, base_dir):
        dataset = statsbomb.load(
            lineup_data=base_dir / "files/statsbomb_lineup.json",
            event_data=base_dir / "files/statsbomb_event.json",
        )
        df = dataset.to_df(engine="pandas")
        incomplete_passes = df[
            (df.event_type == "PASS") & (df.result == "INCOMPLETE")
        ].reset_index()
        assert incomplete_passes.loc[0, "end_coordinates_y"] == pytest.approx(
            0.91519, 1e-4
        )
        assert incomplete_passes.loc[0, "end_coordinates_x"] == pytest.approx(
            0.70945, 1e-4
        )

    def test_to_pandas_additional_columns(self):
        tracking_data = self._get_tracking_dataset()

        data_frame = tracking_data.to_df(
            "*",  # Get all default columns
            match="test",
            bonus_column=lambda frame: frame.frame_id + 10,
            engine="pandas",
        )

        expected_data_frame = DataFrame.from_dict(
            {
                "frame_id": [1, 2],
                "period_id": [1, 2],
                "timestamp": [0.1, 0.2],
                "ball_state": [None, None],
                "ball_owning_team_id": ["home", "away"],
                "ball_x": [100, 0],
                "ball_y": [-50, 50],
                "ball_z": [0, 1],
                "ball_speed": [None, None],
                "match": ["test", "test"],
                "bonus_column": [11, 12],
                "home_1_x": [None, 15],
                "home_1_y": [None, 35],
                "home_1_d": [None, 0.03],
                "home_1_s": [None, 10.5],
                "home_1_extra_data": [None, 1],
                "extra_data": [None, 1],
            }
        )

        assert_frame_equal(data_frame, expected_data_frame, check_like=True)

    def test_event_dataset_to_polars(self, base_dir):
        """
        Make sure an event dataset can be exported as a Polars DataFrame
        """
        dataset = statsbomb.load(
            lineup_data=base_dir / "files/statsbomb_lineup.json",
            event_data=base_dir / "files/statsbomb_event.json",
        )
        df = dataset.to_df(engine="polars")

        import polars as pl

        c = df.select(pl.col("event_id").count())[0, 0]
        assert c == 4061

    def test_tracking_dataset_to_polars(self):
        """
        Make sure a tracking dataset can be exported as a Polars DataFrame
        """
        dataset = self._get_tracking_dataset()

        df = dataset.to_df(engine="polars")

        import polars as pl

        c = df.select(pl.col("frame_id").count())[0, 0]
        assert c == 2

    def test_to_df_config(self):
        """
        Make sure to_df get engine from config. By default, pandas, otherwise polars
        """

        import pandas as pd
        import polars as pl

        dataset = self._get_tracking_dataset()
        df = dataset.to_df()
        assert isinstance(df, pd.DataFrame)

        with config_context("dataframe.engine", "polars"):
            df = dataset.to_df()
            assert isinstance(df, pl.DataFrame)

    @pytest.mark.skipif(sys.version_info < (3, 8), reason="requires python3.8")
    def test_to_df_pyarrow(self):
        """
        Make sure we can export to pandas[pyarrow]. Only works for Python > 3.7.

        The pyarrow engine is part of pandas >=1.5. Pandas 1.3 was the last
        version that supports python 3.7, and does not support pyarrow.
        """
        import pandas as pd

        dataset = self._get_tracking_dataset()
        df = dataset.to_df(engine="pandas[pyarrow]")
        assert isinstance(df, pd.DataFrame)
        assert isinstance(df.dtypes["ball_x"], pd.ArrowDtype)



from dataclasses import dataclass, field
from typing import Any, Callable, Dict, Optional, Union

from kloppy.domain.models.common import DatasetType
from kloppy.utils import (
    deprecated,
    docstring_inherit_attributes,
)

from .common import DataRecord, Dataset, Player
from .pitch import Point, Point3D


@dataclass
class PlayerData:
    coordinates: Point
    distance: Optional[float] = None
    speed: Optional[float] = None
    other_data: Dict[str, Any] = field(default_factory=dict)


@docstring_inherit_attributes(DataRecord)
@dataclass(repr=False)
class Frame(DataRecord):
    """
    Tracking data frame.

    Attributes:
        frame_id: The unique identifier of the frame. Aias for `record_id`.
        ball_coordinates: The coordinates of the ball
        players_data: A dictionary containing the tracking data for each player.
        ball_speed: The speed of the ball
        other_data: A dictionary containing additional data
    """

    frame_id: int
    players_data: Dict[Player, PlayerData]
    other_data: Dict[str, Any]
    ball_coordinates: Point3D
    ball_speed: Optional[float] = None

    @property
    def record_id(self) -> int:
        return self.frame_id

    @property
    def players_coordinates(self):
        return {
            player: player_data.coordinates
            for player, player_data in self.players_data.items()
        }

    def __str__(self):
        return f"<{self.__class__.__name__} frame_id='{self.frame_id}' time='{self.time}'>"

    def __repr__(self):
        return str(self)


@dataclass
@docstring_inherit_attributes(Dataset)
class TrackingDataset(Dataset[Frame]):
    """
    A tracking dataset.

    Attributes:
        dataset_type (DatasetType): `"DatasetType.TRACKING"`
        frames (List[Frame]): A list of frames. Alias for `records`.
        frame_rate (float): The frame rate (in Hertz) at which the data was recorded.
        metadata (Metadata): Metadata of the tracking dataset.
    """

    dataset_type: DatasetType = DatasetType.TRACKING

    @property
    def frames(self):
        return self.records

    @property
    def frame_rate(self):
        return self.metadata.frame_rate

    @deprecated(
        "to_pandas will be removed in the future. Please use to_df instead."
    )
    def to_pandas(
        self,
        record_converter: Optional[Callable[[Frame], Dict]] = None,
        additional_columns=None,
    ) -> "DataFrame":
        try:
            import pandas as pd
        except ImportError:
            raise ImportError(
                "Seems like you don't have pandas installed. Please"
                " install it using: pip install pandas"
            )

        if not record_converter:
            from ..services.transformers.attribute import (
                DefaultFrameTransformer,
            )

            record_converter = DefaultFrameTransformer()

        def generic_record_converter(frame: Frame):
            row = record_converter(frame)
            if additional_columns:
                for k, v in additional_columns.items():
                    if callable(v):
                        value = v(frame)
                    else:
                        value = v
                    row.update({k: value})

            return row

        return pd.DataFrame.from_records(
            map(generic_record_converter, self.records)
        )


__all__ = ["Frame", "TrackingDataset", "PlayerData"]



from math import sqrt
import pytest

from kloppy.domain import (
    Dimension,
    NormalizedPitchDimensions,
    Point,
    Point3D,
    OptaPitchDimensions,
    Unit,
    MetricPitchDimensions,
)
from kloppy.domain.services.transformers import DatasetTransformer


class TestPitchdimensions:
    def test_normalized_pitchdimensions(self):
        pitch_with_scale = NormalizedPitchDimensions(
            x_dim=Dimension(-100, 100),
            y_dim=Dimension(-50, 50),
            pitch_length=120,
            pitch_width=80,
        )

        assert pitch_with_scale.pitch_length == 120
        assert pitch_with_scale.pitch_width == 80

    def test_to_metric_base_dimensions(self):
        pitch = OptaPitchDimensions()

        ifab_point = pitch.to_metric_base(Point(11.5, 50))
        assert ifab_point == Point(11, 34)

        ifab_point = pitch.to_metric_base(Point3D(0, 50, 38))
        assert ifab_point == Point3D(0, 34, 2.44)

        ifab_point = pitch.to_metric_base(
            Point(60, 61), pitch_length=105, pitch_width=68
        )
        assert round(ifab_point.x, 2) == 62.78
        assert round(ifab_point.y, 2) == 41.72

    def test_to_metric_base_dimensions_out_of_bounds(self):
        pitch = NormalizedPitchDimensions(
            x_dim=Dimension(-100, 100),
            y_dim=Dimension(-50, 50),
            pitch_length=120,
            pitch_width=80,
        )
        ifab_point = pitch.to_metric_base(Point(-100, 0))
        assert ifab_point == Point(0, 34)
        ifab_point = pitch.to_metric_base(Point(-105, 0))
        assert ifab_point == Point(-2.625, 34)
        ifab_point = pitch.to_metric_base(Point(105, 0))
        assert ifab_point == Point(107.625, 34)

    def test_from_metric_base_dimensions(self):
        pitch = OptaPitchDimensions()

        opta_point = pitch.from_metric_base(Point(11, 34))
        assert opta_point == Point(11.5, 50)

        opta_point = pitch.from_metric_base(Point3D(0, 34, 2.44))
        assert opta_point == Point3D(0, 50, 38)

        ifab_point = pitch.from_metric_base(
            Point(62.78, 41.72), pitch_length=105, pitch_width=68
        )
        assert round(ifab_point.x, 2) == 60
        assert round(ifab_point.y, 2) == 61

    def test_from_metric_base_dimensions_out_of_bounds(self):
        pitch = NormalizedPitchDimensions(
            x_dim=Dimension(-100, 100),
            y_dim=Dimension(-50, 50),
            pitch_length=120,
            pitch_width=80,
        )
        point = pitch.from_metric_base(Point(0, 34))
        assert point == Point(-100, 0)
        point = pitch.from_metric_base(Point(-2.625, 34))
        assert point == Point(-105, 0)
        point = pitch.from_metric_base(Point(107.625, 34))
        assert point == Point(105, 0)

    def test_distance_between(self):
        pitch = OptaPitchDimensions(pitch_length=105, pitch_width=68)

        distance = pitch.distance_between(
            Point(0, 0),
            Point(100, 100),
        )
        assert distance == sqrt(105**2 + 68**2)

        distance = pitch.distance_between(Point(0, 50), Point(11.5, 50))
        assert distance == 11

        distance = pitch.distance_between(
            Point(100, 50), Point(100 - 11.5, 50)
        )
        assert distance == 11

        distance = pitch.distance_between(
            Point(0, 50), Point(11.5, 50), unit=Unit.CENTIMETERS
        )
        assert distance == 1100

        pitch = NormalizedPitchDimensions(
            x_dim=Dimension(-100, 100),
            y_dim=Dimension(-50, 50),
            pitch_length=120,
            pitch_width=80,
        )
        distance = pitch.distance_between(
            Point(-100, -50),
            Point(100, 50),
        )
        assert distance == sqrt(120**2 + 80**2)

    def test_transform(self):
        transformer = DatasetTransformer(
            from_pitch_dimensions=OptaPitchDimensions(),
            to_pitch_dimensions=MetricPitchDimensions(
                x_dim=Dimension(0, 105),
                y_dim=Dimension(0, 68),
                pitch_length=105,
                pitch_width=68,
                standardized=False,
            ),
        )
        # the corner of the penalty area should remain the corner of
        # the penalty area in the new coordinate system
        transformed_point = transformer.change_point_dimensions(
            Point(17, 78.9)
        )
        assert transformed_point.x == pytest.approx(16.5)
        assert transformed_point.y == pytest.approx(54.16)



from typing import IO, Optional

from .base import OptaEvent, OptaParser
from .f7_xml import F7XMLParser
from .f24_xml import F24XMLParser
from .ma1_json import MA1JSONParser
from .ma1_xml import MA1XMLParser
from .ma3_json import MA3JSONParser
from .ma3_xml import MA3XMLParser


def get_parser(
    feed: IO[bytes],
    feed_code: str,
    feed_format: Optional[str] = None,
    **kwargs,
) -> OptaParser:
    # infer the data format if not provided
    if feed_format is None:
        if feed.read(1).decode("utf-8")[0] == "<":
            feed_format = "XML"
        else:
            feed_format = "JSON"
        feed.seek(0)

    # select the appropriate parser
    feed_code = feed_code.upper()
    feed_format = feed_format.upper()
    if feed_code == "F24" and feed_format == "XML":
        return F24XMLParser(feed)
    elif feed_code == "F7" and feed_format == "XML":
        return F7XMLParser(feed)
    elif feed_code == "MA1" and feed_format == "JSON":
        return MA1JSONParser(feed)
    elif feed_code == "MA1" and feed_format == "XML":
        return MA1XMLParser(feed)
    elif feed_code == "MA3" and feed_format == "JSON":
        return MA3JSONParser(feed)
    elif feed_code == "MA3" and feed_format == "XML":
        return MA3XMLParser(feed)
    else:
        raise NotImplementedError(
            f"A parser for {feed_code} ({feed_format}) feeds is not yet implemented."
        )


__all__ = ["get_parser", "OptaParser", "OptaEvent"]



from kloppy.config import get_config
from kloppy.domain import EventDataset, EventFactory, List, Optional
from kloppy.infra.serializers.event.statsperform import (
    StatsPerformDeserializer,
    StatsPerformInputs,
)
from kloppy.io import FileLike, open_as_file


def load(
    f7_data: FileLike,
    f24_data: FileLike,
    event_types: Optional[List[str]] = None,
    coordinates: Optional[str] = None,
    event_factory: Optional[EventFactory] = None,
) -> EventDataset:
    """
    Load Opta event data.

    Args:
        f7_data: F7 xml feed containing the lineup information.
        f24_data: F24 or F73 xml feed containing the events.
        event_types: A list of event types to load.
        coordinates: The coordinate system to use.
        event_factory: A custom event factory.

    Returns:
        The parsed event data.
    """
    deserializer = StatsPerformDeserializer(
        event_types=event_types,
        coordinate_system=coordinates,
        event_factory=event_factory or get_config("event_factory"),
    )
    with open_as_file(f7_data) as f7_data_fp, open_as_file(
        f24_data
    ) as f24_data_fp:
        return deserializer.deserialize(
            inputs=StatsPerformInputs(
                meta_data=f7_data_fp,
                meta_feed="F7",
                meta_datatype="XML",
                event_data=f24_data_fp,
                event_feed="F24",
                event_datatype="XML",
            ),
        )
